"""Handle argument parsing, defaulting, casting, and more."""

import argparse
import contextlib
import enum
import io
import logging
import re
import sys
import time
import traceback
from typing import Any, Union, cast

import tornado.web
from tornado.escape import to_unicode
from wipac_dev_tools import strtobool

from .handler import RestHandler

LOGGER = logging.getLogger(__name__)


###############################################################################
# Constants

USE_CACHED_VALUE_PLACEHOLDER = "PLACEHOLDER"


class ArgumentSource(enum.Enum):
    """For mapping to argument sources."""

    QUERY_ARGUMENTS = enum.auto()
    JSON_BODY_ARGUMENTS = enum.auto()


###############################################################################
# Utils


def _universal_to_bool(val: Any) -> bool:
    if isinstance(val, bool):
        return val
    elif isinstance(val, str):
        return strtobool(val)
    else:
        return bool(val)


###############################################################################
# Error Patterns

# __main__.py: error: the following arguments are required: --reqd, --bar
ARGUMENTS_REQUIRED_PATTERN = re.compile(r".*(the following arguments are required: .+)")

# __main__.py: error: unrecognized arguments: --xtra 1 --another True False who knows
UNRECOGNIZED_ARGUMENTS_PATTERN = re.compile(r".*(unrecognized arguments:) (.+)")

# argument --foo: invalid int value: 'hank'
INVALID_VALUE_PATTERN = re.compile(r"(argument .+: invalid) .+ value: '.+'")

# argument --pick_it: invalid choice: 'hammer' (choose from 'rock', 'paper', 'scissors')
INVALID_CHOICE_PATTERN = re.compile(r"(argument .+: invalid choice: .+)")

# argument --reco_algo: cannot be empty string / whitespace
FROM_ARGUMENT_TYPE_ERROR_PATTERN = re.compile(r"(argument .+: .+)")


###############################################################################


class ArgumentHandler:
    """Helper class for argument parsing, defaulting, and casting.

    Like argparse.ArgumentParser, but for REST & JSON-body arguments.
    """

    def __init__(
        self, argument_source: ArgumentSource, rest_handler: RestHandler
    ) -> None:
        if sys.version_info < (3, 9):
            # ArgumentParser's `exit_on_error` is only python 3.9+
            self._argparser: argparse.ArgumentParser  # mypy hack
            self.argument_source: ArgumentSource  # mypy hack
            self.rest_handler: RestHandler  # mypy hack
            raise RuntimeError(
                f"{self.__class__.__name__} is supported only for python 3.9+"
            )

        self._argparser = argparse.ArgumentParser(exit_on_error=False)
        self.argument_source = argument_source
        self.rest_handler = rest_handler

    def add_argument(
        self,
        name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Add an argument -- `argparse.add_argument` with minimal additions.

        Do not include a `--`-prefix on the `name`.

        See https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser.add_argument

        ### Built-in Validation
        Validation errors built into `argparse`, such as those triggered
        by the `choices` parameter, are forwarded as `400 Bad Request`
        responses. These responses include the corresponding validation
        message from `argparse`.

        ### `type`-Validation
        Many types of validation exceptions are also handled and returned
        as `400 Bad Request` responses if the `type` parameter is a callable
        that raises an exception.
            > HOWEVER, to include a custom validation message in the
              `400 Bad Request` response, you must raise an
              `argparse.ArgumentTypeError(...)`. Otherwise, the
              `400 Bad Request` response will contain a generic
              message, such as 'argument myarg: invalid type'.

        ### Note
        Not all of `argparse.add_argument`'s parameters make sense
        for key-value based arguments, such as flag-oriented
        options. Nevertheless, no given parameters are excluded,
        just make sure to test it first :)
        """

        def retrieve_json_body_arg(parsed_val: Any) -> Any:
            if parsed_val != USE_CACHED_VALUE_PLACEHOLDER:
                return parsed_val  # this must be the **default** value
            # replace placeholder value with actual value
            key = cast(str, name or kwargs.get("dest"))
            try:
                return self.rest_handler.json_body_arguments[key]
            except KeyError as e:
                # just in case, intercept so the user doesn't get a 400
                raise RuntimeError(
                    f"key '{key}' should exist in json body arguments"
                ) from e

        # TYPE
        if kwargs.get("type") is bool:
            kwargs["type"] = _universal_to_bool
        if self.argument_source == ArgumentSource.JSON_BODY_ARGUMENTS:
            if "type" in kwargs:
                typ = kwargs["type"]  # put in var to avoid unintended recursion
                kwargs["type"] = lambda x: typ(retrieve_json_body_arg(x))
            else:
                kwargs["type"] = retrieve_json_body_arg

        # REQUIRED
        if "default" not in kwargs:
            if "required" not in kwargs:
                # no default? then it's required
                kwargs["required"] = True
            elif not kwargs["required"]:
                raise ValueError(
                    f"Argument '{name}' marked as not required but no default was provided."
                )

        # prepend with '--' and add!
        self._argparser.add_argument(f"--{name}", *args, **kwargs)

    @staticmethod
    def _translate_error(
        exc: Union[Exception, SystemExit],
        captured_stderr: str,
    ) -> str:
        """Translate argparse-style error to a message str for HTTPError."""
        LOGGER.error(f"Intercepted error to translate for requestor -> {exc}")

        # errors not covered by 'exit_on_error=False' (in __init__)
        if isinstance(exc, (SystemExit, argparse.ArgumentError)):
            if isinstance(exc, SystemExit):
                err_msg = captured_stderr
            else:
                err_msg = str(exc)

            # MISSING ARG
            # ex:
            #   b'foo=val'
            #   {'foo': [b'val']}
            #   ['--foo', 'val']
            # stderr:
            #   usage: __main__.py [---h] --reqd REQD --foo FOO --bar BAR
            #   __main__.py: error: the following arguments are required: --reqd, --bar
            if match := ARGUMENTS_REQUIRED_PATTERN.search(err_msg):
                return match.group(1).replace(" --", " ")

            # EXTRA ARG
            # ex:
            #   b'foo=val&reqd=2&xtra=1&another=True&another=False&another=who+knows'
            #   {'foo': [b'val'], 'reqd': [b'2'], 'xtra': [b'1'], 'another': [b'True', b'False', b'who knows']}
            #   ['--foo', 'val', '--reqd', '2', '--xtra', '1', '--another', 'True', 'False', 'who knows']
            # stderr:
            #   usage: __main__.py [---h] --reqd REQD --foo FOO
            #   __main__.py: error: unrecognized arguments: --xtra 1 --another True False who knows
            elif match := UNRECOGNIZED_ARGUMENTS_PATTERN.search(err_msg):
                args = (
                    k.replace("--", "")
                    for k in match.group(2).split()
                    if k.startswith("--")
                )
                return f"{match.group(1)} {', '.join(args)}"

            # INVALID VALUE -- not a system error bc 'exit_on_error=False' (in __init__)
            # ex:
            #   argument --foo: invalid int value: 'hank'
            elif match := INVALID_VALUE_PATTERN.search(err_msg):
                return f"{match.group(1).replace('--', '')} type"
            # ex:
            #   argument --pick_it: invalid choice: 'hammer' (choose from 'rock', 'paper', 'scissors')
            elif match := INVALID_CHOICE_PATTERN.search(err_msg):
                return match.group(1).replace("--", "")

            # argparse.ArgumentTypeError errors -- covers errors in this known format
            # -> this is matched when the server code raises argparse.ArgumentTypeError
            elif match := FROM_ARGUMENT_TYPE_ERROR_PATTERN.search(err_msg):
                return match.group(1).replace("--", "")

        # FALL-THROUGH -- log unknown exception
        ts = time.time()  # log timestamp to aid debugging
        traceback.print_exception(type(exc), exc, exc.__traceback__)
        LOGGER.error(captured_stderr)
        LOGGER.error(type(exc))
        LOGGER.exception(exc)
        LOGGER.error(f"error timestamp: {ts}")
        return f"Unknown argument-handling error ({ts})"

    def parse_args(self) -> argparse.Namespace:
        """Get the args -- like argparse.parse_args but parses a dict."""
        arg_strings: list[str] = []

        # json-encoded body arguments
        if self.argument_source == ArgumentSource.JSON_BODY_ARGUMENTS:
            for key, _ in self.rest_handler.json_body_arguments.items():
                arg_strings.append(f"--{key}")
                # use cached value (see add_argument()) to avoid unneeded encoding & decoding
                arg_strings.append(USE_CACHED_VALUE_PLACEHOLDER)
        # query arguments
        elif self.argument_source == ArgumentSource.QUERY_ARGUMENTS:
            for key, vlist in self.rest_handler.request.arguments.items():
                arg_strings.append(f"--{key}")
                arg_strings.extend(to_unicode(v) for v in vlist)
        # error
        else:
            raise ValueError(f"Invalid argument_source: {self.argument_source}")

        # parse
        with contextlib.redirect_stderr(io.StringIO()) as f:
            try:
                return self._argparser.parse_args(args=arg_strings)
            except (Exception, SystemExit) as e:
                exc = e
                captured_stderr = f.getvalue()
        # handle exception outside of context manager so *this* stderr is not intercepted
        msg = self._translate_error(exc, captured_stderr)
        raise tornado.web.HTTPError(400, reason=msg)



"""Some JSON encoding and decoding utilities."""


# fmt:off
# pylint: skip-file

import ast
import base64
import json
import logging
import zlib
from datetime import date, datetime, time
from typing import Any, Optional, Union

from tornado.escape import recursive_unicode

LOGGER = logging.getLogger(__name__)


JSONType = Any


class json_compressor:
    """Used for files and other large things sent over json.

    Great for log files.
    """
    @staticmethod
    def compress(obj):
        return base64.b64encode(zlib.compress(obj)) if obj else b''

    @staticmethod
    def uncompress(obj):
        return zlib.decompress(base64.b64decode(obj)).decode('utf-8') if obj else ''


class datetime_converter:
    @staticmethod
    def dumps(obj):
        return obj.isoformat()

    @staticmethod
    def loads(obj,name=None):
        if ':' in obj:
            if 'T' in obj or ' ' in obj:
                center = ' '
                if 'T' in obj:
                    center = 'T'
                # must be datetime
                if '.' in obj:
                    return datetime.strptime(obj, "%Y-%m-%d"+center+"%H:%M:%S.%f")
                else:
                    return datetime.strptime(obj, "%Y-%m-%d"+center+"%H:%M:%S")
            else:
                # must be time
                if '.' in obj:
                    return datetime.strptime(obj, "%H:%M:%S.%f")
                else:
                    return datetime.strptime(obj, "%H:%M:%S")
        else:
            # must be date
            return datetime.strptime(obj, "%Y-%m-%d")


class date_converter(datetime_converter):
    @staticmethod
    def loads(obj,name=None):
        d = datetime_converter.loads(obj)
        return date(d.year,d.month,d.day)


class time_converter(datetime_converter):
    @staticmethod
    def loads(obj,name=None):
        d = datetime_converter.loads(obj)
        return time(d.hour,d.minute,d.second,d.microsecond)


class binary_converter:
    """note that is is really only for decode of json, since python bytes are
    strings."""
    @staticmethod
    def dumps(obj,name=None):
        return base64.b64encode(obj)

    @staticmethod
    def loads(obj,name=None):
        return base64.b64decode(obj).decode('utf-8')


class bytearray_converter:
    @staticmethod
    def dumps(obj,name=None):
        return base64.b64encode(str(obj))

    @staticmethod
    def loads(obj,name=None):
        return bytearray(base64.b64decode(obj))


class set_converter:
    @staticmethod
    def dumps(obj):
        return list(obj)

    @staticmethod
    def loads(obj,name=None):
        return set(obj)


# do some default conversions


class repr_converter:
    @staticmethod
    def dumps(obj):
        return repr(obj)

    @staticmethod
    def loads(obj,name=None):
        parts = obj.split('(',1)
        type = parts[0]
        if type not in globals():
            raise Exception()
        parts2 = parts[1].rsplit(')',1)
        args = ast.literal_eval(parts2[0])
        if isinstance(args,tuple):
            ret = globals()['type'](*args)
        else:
            ret = globals()['type'](args)
        return ret


JSONConverters = {
    'datetime':datetime_converter,
    'date':date_converter,
    'time':time_converter,
    'binary':binary_converter,
    'bytearray':bytearray_converter,
    'OrderedDict':repr_converter,
    'set':set_converter,
}


def objToJSON(obj):
    if isinstance(obj,(dict,list,tuple,str,int,float,bool)) or obj is None:
        return obj
    else:
        name = obj.__class__.__name__
        if name in JSONConverters:
            return {'__jsonclass__':[name,JSONConverters[name].dumps(obj)]}
        else:
            LOGGER.error('name: %s, obj: %r', name, obj)
            raise Exception('Cannot encode %s class to JSON'%name)


def JSONToObj(obj):
    ret = obj
    if isinstance(obj,dict) and '__jsonclass__' in obj:
        LOGGER.info('try unpacking class')
        try:
            name = obj['__jsonclass__'][0]
            if name not in JSONConverters:
                raise Exception('class %r not found in converters'%name)
            obj_repr = obj['__jsonclass__'][1]
            ret = JSONConverters[name].loads(obj_repr,name=name)
        except Exception as e:
            LOGGER.warning('error making json class: %r',e,exc_info=True)
    return ret


def json_encode(value: JSONType, indent: Optional[Union[int, str]] = None) -> str:
    """JSON-encodes the given Python object."""
    string = json.dumps(
        recursive_unicode(value),
        default=objToJSON,
        separators=(",", ":"),
        indent=indent,
    )
    return string.replace("</", "<\\/")


def json_decode(value: Union[str, bytes, bytearray]) -> JSONType:
    """Return Python objects for the given JSON string."""
    return json.loads(value, object_hook=JSONToObj)



"""Sub-package __init__."""

from . import utils
from .client import (
    MAX_RETRIES,
    CalcRetryFromBackoffMax,
    CalcRetryFromWaittimeMax,
    RestClient,
)
from .client_credentials import ClientCredentialsAuth
from .device_client import DeviceGrantAuth, SavedDeviceGrantAuth
from .openid_client import OpenIDRestClient
from .session import AsyncSession, Session

__all__ = [
    "RestClient",
    "OpenIDRestClient",
    "ClientCredentialsAuth",
    "DeviceGrantAuth",
    "SavedDeviceGrantAuth",
    "AsyncSession",
    "Session",
    "CalcRetryFromBackoffMax",
    "CalcRetryFromWaittimeMax",
    "MAX_RETRIES",
    "utils",
]



import pytest
import time

from rest_tools.server.session import Session, SessionWrapper, redis_available

STORAGE_TYPES = ['memory']
if redis_available:
    STORAGE_TYPES.append('redis')


@pytest.mark.parametrize('storage_type', STORAGE_TYPES)
def test_memory_session(storage_type):
    session_store = Session(storage_type=storage_type, expiration=0.1)

    # Make sure we start with a clear session storage
    session_store.clear()

    # Create a session
    session_store.set('user123', {'username': 'Alice'})

    # Access the session - this will update the last_accessed time and extend expiry
    time.sleep(.03)
    session_data = session_store.get_session('user123')
    assert session_data == {'username': 'Alice'}

    time.sleep(.08)
    # session should still exist
    session_store.get_session('user123')

    time.sleep(.03)
    # Update the session
    session_store.set('user123', {'email': 'alice@example.com'})

    time.sleep(.08)
    # session should still exist
    session_store.get_session('user123')

    # Let the session expire
    time.sleep(.1)
    with pytest.raises(KeyError):
        session_store.get_session('user123')

    # Manual cleanup
    session_store.set('user123', {'username': 'Alice'})
    time.sleep(.11)
    session_store.cleanup_expired_sessions()
    with pytest.raises(KeyError):
        session_store.get_session('user123')

    # Demonstrate deleting a session
    session_store.set('test_session', {'key': 'value'})
    session_store.get_session('test_session')
    session_store.delete_session('test_session')
    with pytest.raises(KeyError):
        session_store.get_session('test_session')

    session_store.close()


@pytest.mark.parametrize('storage_type', STORAGE_TYPES)
def test_session_wrapper(storage_type):
    session_store = Session(storage_type=storage_type, expiration=0.1)

    session_store.set('user123', {'username': 'Alice'})

    data = session_store.get_session('user123')

    data_wrapper = SessionWrapper(key='user123', data=data, session_mgr=session_store)

    time.sleep(.05)
    # should reset expiration
    data_wrapper['foo'] = 'bar'

    time.sleep(.07)
    ret = session_store.get_session('user123')
    assert ret == {'username': 'Alice', 'foo': 'bar'}

    # should reset expiration
    del data_wrapper['foo']

    time.sleep(.07)
    ret = session_store.get_session('user123')
    assert ret == {'username': 'Alice'}

    session_store.close()



"""decorators.py."""

# fmt:off

import logging
import os
import re
from functools import wraps
from inspect import isawaitable
from typing import Protocol

import requests.exceptions
import tornado.web

from .. import telemetry as wtt

LOGGER = logging.getLogger(__name__)


########################################################################################################################


def authenticated(method):
    """Decorate methods with this to require that the Authorization header is
    filled with a valid token. Does *not* check the authorization of the token,
    just that it exists.

    On failure, raises a 403 error.

    Raises:
        :py:class:`tornado.web.HTTPError`
    """

    @wraps(method)
    async def wrapper(self, *args, **kwargs):
        if not self.current_user:
            raise tornado.web.HTTPError(403, reason="authentication failed")
        ret = method(self, *args, **kwargs)
        if isawaitable(ret):
            return await ret
        else:
            return ret
    return wrapper


########################################################################################################################


def catch_error(method):
    """Decorator to catch and handle errors on handlers.

    All failures caught here
    """
    @wraps(method)
    async def wrapper(self, *args, **kwargs):
        try:
            ret = method(self, *args, **kwargs)
            if isawaitable(ret):
                return await ret
            else:
                return ret
        except tornado.web.HTTPError:
            raise  # tornado can handle this
        except tornado.httpclient.HTTPError:
            raise  # tornado can handle this
        except requests.exceptions.HTTPError as e:
            LOGGER.warning('Error in website handler', exc_info=True)
            try:
                self.statsd.incr(self.__class__.__name__+'.error')
            except Exception:
                pass  # ignore statsd errors
            if e.response.status_code == 403:
                code = 403
                message = 'Error authenticating user'
            else:
                code = 500
                message = 'Error contacting backend in '+self.__class__.__name__
            self.send_error(code, reason=message)
        except Exception:
            LOGGER.warning('Error in website handler', exc_info=True)
            try:
                self.statsd.incr(self.__class__.__name__+'.error')
            except Exception:
                pass  # ignore statsd errors
            message = 'Error in '+self.__class__.__name__
            self.send_error(500, reason=message)
        return None
    return wrapper


########################################################################################################################


def role_authorization(**_auth):
    """Handle RBAC authorization.

    Like :py:func:`authenticated`, this requires the Authorization header
    to be filled with a valid token.  Note that calling both decorators
    is not necessary, as this decorator will perform authentication
    checking as well.

    Args:
        roles (list): The roles to match

    Raises:
        :py:class:`tornado.web.HTTPError`
    """
    def make_wrapper(method):
        @authenticated
        @catch_error
        @wraps(method)
        async def wrapper(self, *args, **kwargs):
            roles = _auth.get('roles', [])

            auth_role = self.auth_data.get('role',None)
            if roles and auth_role in roles:
                wtt.set_current_span_attribute('self.auth_data.roles', auth_role)
            else:
                LOGGER.info('roles: %r', roles)
                LOGGER.info('token_role: %r', auth_role)
                LOGGER.info('role mismatch')
                raise tornado.web.HTTPError(403, reason="authorization failed")

            ret = method(self, *args, **kwargs)
            if isawaitable(ret):
                return await ret
            else:
                return ret
        return wrapper
    return make_wrapper


########################################################################################################################


def scope_role_auth(**_auth):
    """Handle RBAC authorization using oauth2 scopes.

    Like :py:func:`authenticated`, this requires the Authorization header
    to be filled with a valid token.  Note that calling both decorators
    is not necessary, as this decorator will perform authentication
    checking as well.

    Args:
        roles (list): The roles to match
        prefix (str): The scope prefix

    Raises:
        :py:class:`tornado.web.HTTPError`
    """
    def make_wrapper(method):
        @authenticated
        @catch_error
        @wraps(method)
        async def wrapper(self, *args, **kwargs):
            roles = _auth.get('roles', [])
            scope_prefix = _auth.get('prefix', None)

            authorized = False

            auth_roles = []
            for scope in self.auth_data.get('scope', '').split():
                if scope_prefix and scope.startswith(f'{scope_prefix}:'):
                    auth_roles.append(scope.split(':', 1)[-1])

            authorized = set(roles).intersection(auth_roles)

            if not authorized:
                LOGGER.info('roles: %r', roles)
                LOGGER.info('token_roles: %r', auth_roles)
                LOGGER.info('role mismatch')
                raise tornado.web.HTTPError(403, reason="authorization failed")

            wtt.set_current_span_attribute('self.auth_data.roles', ','.join(sorted(authorized)))

            ret = method(self, *args, **kwargs)
            if isawaitable(ret):
                return await ret
            else:
                return ret
        return wrapper
    return make_wrapper


########################################################################################################################


def keycloak_role_auth(**_auth):
    """Handle RBAC authorization using keycloak realm roles.
    Like :py:func:`authenticated`, this requires the Authorization header
    to be filled with a valid token.  Note that calling both decorators
    is not necessary, as this decorator will perform authentication
    checking as well.

    Args:
        roles (list): The roles to match
        prefix (str): The token prefix (default: realm_access.roles)
    Raises:
        :py:class:`tornado.web.HTTPError`
    """
    def make_wrapper(method):
        @authenticated
        @catch_error
        @wraps(method)
        async def wrapper(self, *args, **kwargs):
            roles = _auth.get('roles', [])
            prefix = _auth.get('prefix', 'realm_access.roles').split('.')

            auth_roles = self.auth_data
            while prefix:
                auth_roles = auth_roles.get(prefix[0], {})
                prefix = prefix[1:]

            authorized = set(roles).intersection(auth_roles)

            if not authorized:
                LOGGER.info('roles: %r', roles)
                LOGGER.info('token_roles: %r', auth_roles)
                LOGGER.info('role mismatch')
                raise tornado.web.HTTPError(403, reason="authorization failed")

            wtt.set_current_span_attribute('self.auth_data.roles', ','.join(sorted(authorized)))

            ret = method(self, *args, **kwargs)
            if isawaitable(ret):
                return await ret
            else:
                return ret
        return wrapper
    return make_wrapper


########################################################################################################################


def token_attribute_role_mapping_auth(role_attrs, group_attrs=None):  # noqa: C901
    """Handle RBAC authorization by creating a decorator that maps token
    attributes to roles, then using the decorator to allow access to functions
    by role.  Can also map groups and add that to auth data.

    Like :py:func:`authenticated`, this requires the Authorization header
    to be filled with a valid token.  Note that calling both decorators
    is not necessary, as this decorator will perform authentication
    checking as well.

    Use an attr expression for evaluating a role or group.  Some examples:
      * For an exact match on a string: `attr=foo`
      * For a list, the element must exist in the list: `my.attr=foo`
      * Scope is handled like a list, splitting the string on spaces

    Python regular expressions are supported on values, which can also be used
    to capture a value to use as the name of the role or group:
      * For matching any value: `my.attr=.*`
      * For capturing a value: `my.attr=(foo)`
      * For capturing part of a value: `my.attr=foo(.*)`
    Note that this uses `re.fullmatch`.

    Args:
        role_attrs (dict): Map of role name to list of valid attr expressions
        group_attrs (dict): Map of group name to list of valid attr expressions
    Returns:
        callable: Decorator function

    Decorator Args:
        roles (list): The roles to match
    Decorator Raises:
        :py:class:`tornado.web.HTTPError`

    Example:
        my_auth = token_attribute_role_mapping_auth(
            role_attrs = {
                'write': ['groups=my-service-write', 'groups=admins'],
                'read': ['groups'],
            },
            group_attrs = {
                r'\1': [r'groups=my-service-(.*)']
            }
        )

        class Handler:
            @my_auth(roles=['read', 'write'])
            async def get(self):
                pass
            @my_auth(roles['write'])
            async def post(self):
                pass
    """
    basic_role_attrs = set()
    regex_role_attrs = set()
    for name in role_attrs:
        if re.fullmatch(r'\w+', name):
            basic_role_attrs.add(name)
        else:
            regex_role_attrs.add(name)
    if not group_attrs:
        group_attrs = {}

    def eval_expression(token, e):
        name, val = e.split('=',1)
        if name == 'scope':
            # special handling to split into string
            token_val = token.get('scope','').split()
        else:
            prefix = name.split('.')[:-1]
            while prefix:
                token = token.get(prefix[0], {})
                prefix = prefix[1:]
            token_val = token.get(name.split('.')[-1], None)

        LOGGER.debug('token_val = %r', token_val)
        if token_val is None:
            return []

        prog = re.compile(val)
        if isinstance(token_val, list):
            ret = (prog.fullmatch(v) for v in token_val)
        else:
            ret = [prog.fullmatch(token_val)]
        return [r for r in ret if r]

    def make_decorator(**_auth):
        def make_wrapper(method):
            @authenticated
            @catch_error
            @wraps(method)
            async def wrapper(self, *args, **kwargs):
                roles = _auth.get('roles', [])

                try:
                    authorized_roles = set()
                    for role in basic_role_attrs.intersection(roles):
                        for expression in role_attrs[role]:
                            if eval_expression(self.auth_data, expression):
                                authorized_roles.add(role)
                    for name in regex_role_attrs:
                        for expression in role_attrs[name]:
                            ret = eval_expression(self.auth_data, expression)
                            rolenames = [match.expand(name) for match in ret]
                            authorized_roles.update(role for role in roles if role in rolenames)
                except Exception as exc:
                    LOGGER.warning('exception in role auth', exc_info=True)
                    raise tornado.web.HTTPError(500, reason="internal server error") from exc

                if not authorized_roles:
                    LOGGER.debug('roles requested: %r', roles)
                    LOGGER.debug('role mismatch')
                    raise tornado.web.HTTPError(403, reason="authorization failed")

                LOGGER.debug('roles requested: %r', roles)
                authorized_roles = sorted(authorized_roles)
                LOGGER.debug('roles authorized: %r', authorized_roles)
                wtt.set_current_span_attribute('self.auth_data.roles', ','.join(authorized_roles))
                self.auth_roles = authorized_roles

                try:
                    authorized_groups = set()
                    for name in group_attrs:
                        for expression in group_attrs[name]:
                            ret = eval_expression(self.auth_data, expression)
                            authorized_groups.update(match.expand(name) for match in ret)
                except Exception as exc:
                    LOGGER.warning('exception in group auth', exc_info=True)
                    raise tornado.web.HTTPError(500, reason="internal server error") from exc

                if authorized_groups:
                    authorized_groups = sorted(authorized_groups)
                    LOGGER.debug('groups authorized: %r', authorized_groups)
                    wtt.set_current_span_attribute('self.auth_data.groups', ','.join(authorized_groups))
                    self.auth_groups = authorized_groups

                ret = method(self, *args, **kwargs)
                if isawaitable(ret):
                    return await ret
                else:
                    return ret
            return wrapper
        return make_wrapper
    return make_decorator


class TokenAttributeRoleMappingProtocol(Protocol):
    auth_groups: list
    auth_roles: list


########################################################################################################################
# fmt:on

try:
    import openapi_core
    from openapi_core import OpenAPI
    from openapi_core.contrib import requests as openapi_core_requests
    from openapi_core.validation.exceptions import ValidationError
    import openapi_core.validation.schemas.exceptions

    openapi_available = True
except ImportError:
    # if client code wants to use these features, then let the built-in errors raise
    openapi_available = False


def validate_request(openapi_spec: "OpenAPI"):  # type: ignore
    """Validate request obj against the given OpenAPI spec."""
    if not openapi_available:
        raise RuntimeError(
            "openapi cannot be imported! perhaps you meant to pip install it?"
        )

    def make_wrapper(method):  # type: ignore[no-untyped-def]
        async def wrapper(zelf: tornado.web.RequestHandler, *args, **kwargs):  # type: ignore[no-untyped-def]
            LOGGER.debug("validating with openapi spec")
            # NOTE - don't change data (unmarshal) b/c we are downstream of data separation
            try:
                # https://openapi-core.readthedocs.io/en/latest/validation.html
                openapi_spec.validate_request(
                    _http_server_request_to_openapi_request(zelf.request),
                )
            except ValidationError as e:  # type: ignore
                LOGGER.error(f"invalid request: {e.__class__.__name__} - {e}")
                if isinstance(  # look at the ORIGINAL exception that caused this error
                    e.__context__,
                    openapi_core.validation.schemas.exceptions.InvalidSchemaValue,  # type: ignore
                ):
                    reason = "; ".join(  # to client
                        # verbose details after newline
                        str(x).split("\n", maxsplit=1)[0]
                        for x in e.__context__.schema_errors
                    )
                else:
                    reason = str(e)  # to client
                if os.getenv("CI"):
                    # in prod, don't fill up logs w/ traces from invalid data
                    LOGGER.exception(e)
                raise tornado.web.HTTPError(
                    status_code=400,
                    log_message=f"{e.__class__.__name__}: {e}",  # to stderr
                    reason=reason,  # to client
                )
            except Exception as e:
                LOGGER.error(f"unexpected exception: {e.__class__.__name__} - {e}")
                LOGGER.exception(e)
                raise tornado.web.HTTPError(
                    status_code=400,
                    log_message=f"{e.__class__.__name__}: {e}",  # to stderr
                    reason=None,  # to client (don't send possibly sensitive info)
                )

            return await method(zelf, *args, **kwargs)

        return wrapper

    return make_wrapper


def _http_server_request_to_openapi_request(
    req: tornado.httputil.HTTPServerRequest,
) -> "openapi_core_requests.RequestsOpenAPIRequest":
    """Convert a `tornado.httputil.HTTPServerRequest` to openapi's type."""
    if not openapi_available:
        raise RuntimeError(
            "openapi cannot be imported! perhaps you meant to pip install it?"
        )
    return openapi_core_requests.RequestsOpenAPIRequest(  # type: ignore
        requests.Request(
            method=req.method.lower() if req.method else "get",
            url=f"{req.protocol}://{req.host}{req.uri}",
            headers=req.headers,
            files=req.files,
            data=req.body if not req.body_arguments else None,  # see below
            params=req.query_arguments,
            # auth=None,
            cookies=req.cookies,
            # hooks=None,
            json=req.body_arguments if req.body_arguments else None,  # see above
        )
    )


########################################################################################################################



import argparse
import asyncio
import logging

from rest_tools.client import SavedDeviceGrantAuth
from rest_tools.client.openid_client import RegisterOpenIDClient


async def get_token(address, scope=None):
    scopes = scope if scope else []
    async with RegisterOpenIDClient(address, 'test') as (client_id, client_secret):
        rest_client = SavedDeviceGrantAuth(
            address='',
            token_url=address,
            filename='token-issuer-refresh-token',
            client_id=client_id,
            client_secret=client_secret,
            scopes=scopes,
        )
        # could use `rest_client` as a `RestClient`, but we're only interested in the token
        return rest_client._openid_token()


def main():
    parser = argparse.ArgumentParser(description='Get an OAuth2 token via client credentials')
    parser.add_argument('--address', default='https://keycloak.icecube.wisc.edu/auth/realms/IceCube', help='OAuth2 server address')
    parser.add_argument('--scope', action='append', help='optional scopes')
    parser.add_argument('--log-level', default='info', help='logging level')

    args = parser.parse_args()
    kwargs = vars(args)

    level = kwargs.pop('log_level').upper()
    logging.basicConfig(level=getattr(logging, level))

    logging.info('scopes: %r', kwargs['scope'])

    print(asyncio.run(get_token(**kwargs)))


if __name__ == '__main__':
    main()



"""RestHandler and related classes."""

import base64
from collections.abc import Callable, Awaitable
import functools
import hmac
from inspect import isawaitable
import json
import logging
import time
import urllib.parse
from collections import defaultdict
from typing import Any, Optional, Union

import tornado.escape
import tornado.httputil
import tornado.web
from tornado.auth import OAuth2Mixin

from .decorators import catch_error
from .stats import RouteStats
from .. import telemetry as wtt
from ..utils.auth import Auth, OpenIDAuth
from ..utils.json_util import json_decode
from ..utils.pkce import PKCEMixin

LOGGER = logging.getLogger(__name__)


# fmt:off


def _log_auth_failed(e: Exception):
    LOGGER.info('failed auth')
    if LOGGER.isEnabledFor(logging.DEBUG):
        LOGGER.exception(e)


def RestHandlerSetup(config={}):
    """Default RestHandler setup. Returns the default arguments for passing to
    the route setup.

    Args:
        config (dict): config dict

    Returns:
        dict: handler config
    """
    debug = True if 'debug' in config and config['debug'] else False
    auth = None
    auth_url = ''
    module_auth_key = ''
    if 'auth' in config:
        kwargs = {}
        if 'leeway' in config['auth']:
            kwargs['leeway'] = config['auth']['leeway']
        if 'audience' in config['auth']:
            kwargs['audience'] = config['auth']['audience']
        if 'issuers' in config['auth']:
            kwargs['issuers'] = config['auth']['issuers']
        if 'secret' in config['auth']:
            kwargs['secret'] = config['auth']['secret']
            if 'issuer' in config['auth']:
                kwargs['issuer'] = config['auth']['issuer']
            if 'algorithm' in config['auth']:
                kwargs['algorithm'] = config['auth']['algorithm']
            if 'expiration' in config['auth']:
                kwargs['expiration'] = config['auth']['expiration']
            if 'expiration_temp' in config['auth']:
                kwargs['expiration_temp'] = config['auth']['expiration_temp']
            auth = Auth(**kwargs)
        elif 'openid_url' in config['auth']:
            if 'algorithms' in config['auth']:
                kwargs['algorithms'] = config['auth']['algorithms']
            auth = OpenIDAuth(config['auth']['openid_url'], **kwargs)
            if auth.token_url:
                auth_url = auth.token_url
        if 'url' in config['auth']:
            auth_url = config['auth']['url']
    if 'rest_api' in config and 'auth_key' in config['rest_api']:
        module_auth_key = config['rest_api']['auth_key']

    if 'route_stats' in config:
        route_stats = defaultdict(functools.partial(RouteStats, **config['route_stats']))
    else:
        route_stats = defaultdict(RouteStats)

    return {
        'debug': debug,
        'auth': auth,
        'auth_url': auth_url,
        'module_auth_key': module_auth_key,
        'server_header': config.get('server_header', 'REST'),
        'route_stats': route_stats
    }


class RestHandler(tornado.web.RequestHandler):
    """Default REST handler."""
    def __init__(self, *args, **kwargs) -> None:
        self.server_header = ''
        try:
            super().__init__(*args, **kwargs)
        except Exception:
            LOGGER.error('error', exc_info=True)

    def initialize(self, debug=False, auth: Union[Auth, None] = None, auth_url=None, module_auth_key='', server_header='', route_stats=None, **kwargs):
        super().initialize(**kwargs)
        self.debug = debug
        self.auth = auth
        self.auth_url = auth_url
        self.auth_data: dict[str, Any] = {}
        self.auth_key = ''
        self.module_auth_key = module_auth_key
        self.server_header = server_header
        self.route_stats = route_stats

    @wtt.spanned(
        span_namer=wtt.SpanNamer(use_this_arg='self.request.method'),
        kind=wtt.SpanKind.SERVER,
        these=["self.request.method", "self.request.path"],
        carrier="self.request.headers",
    )
    async def _execute(self, *args: Any, **kwargs: Any) -> None:
        """Call implemented methods.

        NOTE: This is the closest common call-stack ancestor of
            - `prepare()`,
            - "method handlers" (`get()`, `post()`, ...),
            - `on_finish()`,
            - etc.
        """
        return await super()._execute(*args, **kwargs)

    def set_default_headers(self):
        self._headers['Server'] = self.server_header

    def get_current_user(self):
        """Get the current user, and set auth-related attributes."""
        try:
            type, token = self.request.headers['Authorization'].split(' ', 1)
            if type.lower() != 'bearer':
                raise Exception('bad header type')
            LOGGER.debug('token: %r', token)
            if self.auth is None:
                raise RuntimeError('need to set auth in order to validate tokens!')
            data = self.auth.validate(token)
            self.auth_data = data
            self.auth_key = token
            return data['sub']
        # Auth Failed
        except Exception as e:
            if self.debug and 'Authorization' in self.request.headers:
                LOGGER.info('Authorization: %r', self.request.headers['Authorization'])
            _log_auth_failed(e)

        return None

    @wtt.evented()
    def prepare(self):
        """Prepare before http-method request handlers."""

        # log at the very start of every new request
        logging.info(">>> %s", self._request_summary())  # use the root logger
        # ^^^ this mimics the log line that tornado provides at the end of a request:
        #       see rest_tools.server.tornado_logger().
        #       ">>>" makes logs line-up (end-of-request log line uses http code: 200, 400, etc.)
        LOGGER.debug(f"[{self.__class__.__name__}]")

        if self.route_stats is not None:
            stat = self.route_stats[self.request.path]
            if stat.is_overloaded():
                backoff = stat.get_backoff_time()
                LOGGER.warning('Server is overloaded, backoff %r', backoff)
                self.set_header('Retry-After', backoff)
                raise tornado.web.HTTPError(503, reason="server overloaded")
            self.start_time = time.time()

    @wtt.evented()
    def on_finish(self):
        """Cleanup after http-method request handlers."""
        if self.route_stats is not None and self.get_status() < 500:
            stat = self.route_stats[self.request.path]
            stat.append(time.time() - self.start_time)

    @wtt.evented(all_args=True)
    def write_error(self, status_code=500, **kwargs):
        """Write out custom error json."""
        data = {
            'code': status_code,
            'error': self._reason,
        }
        self.write(data)
        self.finish()

    @functools.cached_property
    def json_body_arguments(self) -> dict[str,Any]:
        """Get the body arguments, decoded from a JSON-encoded request body."""
        if not self.request.body:
            return {}

        try:
            args = json_decode(self.request.body)
        except json.JSONDecodeError:
            raise tornado.web.HTTPError(
                400, reason="requests body is not JSON-encoded"
            )

        if not isinstance(args, dict):
            raise tornado.web.HTTPError(
                400, reason="JSON-encoded requests body must be a 'dict'"
            )
        return args

    def get_argument(
        self,
        name: str,
        default: Any = tornado.web._ARG_DEFAULT,
        strip: bool = True,
        # deprecated args
        type: Any = None,
        choices: Any = None,
        forbiddens: Any = None,
        strict_type: Any = None,
    ) -> Any:
        """Get argument from query base-arguments / JSON-decoded request-body.

        If no `default` is provided, and the argument is not present, raise `400`.

        Arguments:
            name -- the argument's name

        Keyword Arguments:
            default -- a default value to use if the argument is not present
            strip {`bool`} -- whether to `str.strip()` the arg's value (default: {`True`})
            type -- **deprecated**
            choices -- **deprecated**
            forbiddens -- **deprecated**
            strict_type -- **deprecated**

        Returns:
            the argument's value, possibly stripped
        """
        if type is not None or choices is not None or forbiddens is not None or strict_type is not None:
            raise RuntimeError("advanced argument checking is deprecated, please use ArgumentHandler instead")

        # 1st: check query args w/o default
        try:
            return super().get_argument(name, strip=strip)  # no default
        except tornado.web.MissingArgumentError:
            pass

        # 2nd: check json-body args w/o default
        try:
            return self.json_body_arguments[name]  # no default
        except KeyError:
            pass

        # Finally: check json-body args w/ default
        return super().get_argument(name, default, strip=strip)


class KeycloakUsernameMixin:
    """Get the username correctly from Keycloak tokens.

    Note: will not work on service account tokens.
    """
    auth_data: dict

    def get_current_user(self):
        if not super().get_current_user():  # type: ignore
            return None
        username = self.auth_data.get('preferred_username', None)
        if not username:
            username = self.auth_data.get('upn', None)
            if not username:
                LOGGER.info('could not find auth username')
        return username


class OpenIDCookieHandlerMixin:
    """Store/load current user's `OpenIDLoginHandler` tokens in cookies."""
    auth: Union[Auth, None]
    get_secure_cookie: Callable[..., Optional[bytes]]
    set_secure_cookie: Callable[..., None]
    clear_cookie: Callable[..., None]

    def get_current_user(self) -> Union[str, None]:
        """Get the current user, and set auth-related attributes."""
        try:
            access_token = self.get_secure_cookie('access_token')
            if not access_token:
                raise RuntimeError('missing access_token cookie')
            if not self.auth:
                raise RuntimeError('auth is not set!')
            data = self.auth.validate(access_token)
            self.auth_data = data
            self.auth_key = access_token.decode('utf-8')
            refresh_token = self.get_secure_cookie('refresh_token')
            self.auth_refresh_token = refresh_token
            return data['sub']
        # Auth Failed
        except Exception as e:
            _log_auth_failed(e)
        return None

    def store_tokens(
        self,
        access_token,
        access_token_exp,
        refresh_token=None,
        refresh_token_exp=None,
        user_info=None,
        user_info_exp=None,
    ) -> Union[None, Awaitable[None]]:
        """Store jwt tokens and user info from OpenID-compliant auth source.

        Args:
            access_token (str): jwt access token
            access_token_exp (int): access token expiration in seconds
            refresh_token (str): jwt refresh token
            refresh_token_exp (int): refresh token expiration in seconds
            user_info (dict): user info (from id token or user info lookup)
            user_info_exp (int): user info expiration in seconds
        """
        self.set_secure_cookie('access_token', access_token,
                               expires_days=float(access_token_exp)/3600/24)
        if refresh_token and refresh_token_exp:
            self.set_secure_cookie('refresh_token', refresh_token,
                                   expires_days=float(refresh_token_exp)/3600/24)
        if user_info and user_info_exp:
            self.set_secure_cookie('user_info', tornado.escape.json_encode(user_info),
                                   expires_days=float(user_info_exp)/3600/24)
        return None

    def clear_tokens(self):
        """Clear token data, usually on logout."""
        self.clear_cookie('access_token')
        self.clear_cookie('refresh_token')
        self.clear_cookie('user_info')


class OpenIDLoginHandler(OpenIDCookieHandlerMixin, OAuth2Mixin, PKCEMixin, RestHandler):
    """Handle OpenID Connect logins.

    Should be combined with an appropriate mixin to store the token(s).
    `OpenIDCookieHandlerMixin` is used by default, but can be overridden.

    Requires the `login_url` application setting to be a full url.
    """
    store_tokens: Callable[..., Awaitable[None]]

    def initialize(self, oauth_client_id, oauth_client_secret, oauth_client_scope=None, **kwargs):  # type: ignore
        super().initialize(**kwargs)
        if not isinstance(self.auth, OpenIDAuth):
            raise RuntimeError('OpenID Connect auth not set up')
        self._OAUTH_AUTHORIZE_URL = self.auth.provider_info['authorization_endpoint']
        self._OAUTH_ACCESS_TOKEN_URL = self.auth.provider_info['token_endpoint']
        self._OAUTH_LOGOUT_URL = self.auth.provider_info['end_session_endpoint']
        self._OAUTH_USERINFO_URL = self.auth.provider_info['userinfo_endpoint']
        self.oauth_client_id = oauth_client_id
        self.oauth_client_secret = oauth_client_secret

        scopes = {'openid', 'profile'}
        if oauth_client_scope is not None:
            scopes.update(oauth_client_scope.split())
        else:
            scopes.add('groups')
        if oauth_client_secret:
            scopes.add('offline_access')
        self.oauth_client_scope = list(scopes)

    async def get_authenticated_user(
        self, redirect_uri: str, code: str, state: dict[str, Any]
    ) -> dict[str, Any]:
        http = self.get_auth_http_client()
        body = {
            'redirect_uri': redirect_uri,
            'code': code,
            'client_id': self.oauth_client_id,
            'grant_type': 'authorization_code',
        }
        kwargs = {}
        if self.oauth_client_secret:
            kwargs['auth_username'] = self.oauth_client_id
            kwargs['auth_password'] = self.oauth_client_secret
        else:
            # use PKCE
            code_challenge = state.get('code_challenge', None)
            if not code_challenge:
                raise tornado.web.HTTPError(400, reason='missing PKCE code challenge')
            body['code_verifier'] = self.get_pkce_verifier(code_challenge)

        response = await http.fetch(
            self._OAUTH_ACCESS_TOKEN_URL,
            method='POST',
            headers={'Content-Type': 'application/x-www-form-urlencoded'},
            body=urllib.parse.urlencode(body),
            **kwargs,
        )
        ret = tornado.escape.json_decode(response.body)
        if not ret.get('id_token', ''):
            response = await http.fetch(
                self._OAUTH_USERINFO_URL,
                method='GET',
                headers={'Authorization': f'Bearer {ret["access_token"]}'},
            )
            ret['id_token'] = tornado.escape.json_decode(response.body)

        try:
            if not self.auth:
                raise RuntimeError('auth is not set!')
            if ret.get('id_token') and isinstance(ret['id_token'], str):
                ret['id_token'] = self.auth.validate(ret['id_token'])
            self.auth.validate(ret['access_token'])
        except Exception:
            if self.debug:
                LOGGER.debug(f'bad token: {ret}')
            raise
        return ret

    def _decode_state(self, state: Union[bytes, str]) -> dict[str, Any]:
        data = tornado.escape.json_decode(base64.b64decode(state))
        _, token, _ = self._decode_xsrf_token(data.pop('xsrf'))
        _, expected_token, _ = self._get_raw_xsrf_token()
        if not token:
            raise tornado.web.HTTPError(403, reason="'state' argument has invalid format")
        if not hmac.compare_digest(tornado.escape.utf8(token), tornado.escape.utf8(expected_token)):
            raise tornado.web.HTTPError(403, reason="XSRF cookie does not match state argument")
        return data

    def _encode_state(self, data: dict[str, Any]) -> bytes:
        data2 = data.copy()  # make a copy to not add xsrf to source dict
        data2['xsrf'] = self.xsrf_token.decode('utf-8')
        return base64.b64encode(tornado.escape.json_encode(data2).encode('utf-8'))

    @catch_error
    async def get(self):
        if self.get_argument('error', False):
            err = self.get_argument('error_description', None)
            if not err:
                err = 'unknown oauth2 error'
            raise tornado.web.HTTPError(400, reason=err)
        elif self.get_argument('code', False):
            data = self._decode_state(self.get_argument('state'))
            user = await self.get_authenticated_user(
                redirect_uri=self.get_login_url(),
                code=self.get_argument('code'),
                state=data,
            )

            # set expire times (can be 0 in user data, which is an invalid cookie)
            access_expire = user.get('expires_in', 0)
            if not access_expire:
                access_expire = 1800
            refresh_expire = user.get('refresh_expires_in', 0)
            if not refresh_expire:
                refresh_expire = 86400

            # Save the user data
            ret = self.store_tokens(
                access_token=user['access_token'],
                access_token_exp=access_expire,
                refresh_token=user.get('refresh_token'),
                refresh_token_exp=refresh_expire,
                user_info=user.get('id_token'),
                user_info_exp=refresh_expire if user.get('refresh_token') else access_expire,
            )
            if ret and isawaitable(ret):
                await ret

            if data.get('redirect', None):
                url = data['redirect']
                if 'state' in data:
                    url = tornado.httputil.url_concat(url, {'state': data['state']})
                self.redirect(url)
            elif self.settings.get('debug', False):
                self.write(user)
            else:
                raise tornado.web.HTTPError(400, reason='missing redirect')
        else:
            state = {}
            redirect = self.get_argument('next', None)
            if not redirect:
                redirect = self.get_argument('redirect', None)
            if redirect:
                state['redirect'] = redirect
            elif not self.settings.get('debug', False):
                raise tornado.web.HTTPError(400, 'missing redirect')
            if self.get_argument('state', False):
                state['state'] = self.get_argument('state')
            extra_params = {}
            if not self.oauth_client_secret:
                # use PKCE  https://www.rfc-editor.org/rfc/rfc7636
                extra_params['code_challenge'] = self.create_pkce_challenge()
                extra_params['code_challenge_method'] = 'S256'
                state['code_challenge'] = extra_params['code_challenge']
            extra_params["state"] = self._encode_state(state)

            self.authorize_redirect(
                redirect_uri=self.get_login_url(),
                client_id=self.oauth_client_id,
                scope=self.oauth_client_scope,
                extra_params=extra_params,
                response_type='code')



import argparse
import logging

from rest_tools.client import SavedDeviceGrantAuth


def get_token(address, client_id, client_secret=None, scope=None):
    scopes = scope if scope else []
    rest_client = SavedDeviceGrantAuth(
        address='',
        token_url=address,
        filename='device-refresh-token',
        client_id=client_id,
        client_secret=client_secret,
        scopes=scopes,
    )
    # could use `rest_client` as a `RestClient`, but we're only interested in the token
    return rest_client._openid_token()


def main():
    parser = argparse.ArgumentParser(description='Get an OAuth2 token via client credentials')
    parser.add_argument('--address', default='https://keycloak.icecube.wisc.edu/auth/realms/IceCube', help='OAuth2 server address')
    parser.add_argument('--client_secret', default=None, help='client secret')
    parser.add_argument('--scope', action='append', help='optional scopes')
    parser.add_argument('--log-level', default='info', help='logging level')
    parser.add_argument('client_id', help='client id')

    args = parser.parse_args()
    kwargs = vars(args)

    level = kwargs.pop('log_level').upper()
    logging.basicConfig(level=getattr(logging, level))

    print('access token:', get_token(**kwargs))


if __name__ == '__main__':
    main()



"""Sub-package __init__."""

from . import json_util
from .auth import Auth, OpenIDAuth
from .config import from_environment
from .daemon import Daemon

__all__ = [
    "json_util",
    "Auth",
    "OpenIDAuth",
    "Daemon",
    "from_environment",
]



# version is a human-readable version number.

from . import client, server, utils

__all__ = [
    "client",
    "server",
    "utils",
]

# NOTE: `__version__` is not defined because this package is built using 'setuptools-scm' --
#   use `importlib.metadata.version(...)` if you need to access version info at runtime.



"""Test server/arghandler.py."""

import argparse
import json
import re
import sys
from typing import Any, Union, cast
from unittest.mock import Mock

import pytest
import requests
import tornado.web
from tornado import httputil
from wipac_dev_tools import strtobool

from rest_tools.server.arghandler import ArgumentHandler, ArgumentSource
from rest_tools.server.handler import RestHandler

# pylint: disable=W0212,W0621

# these tests are only for 3.9+
if sys.version_info < (3, 9):
    pytest.skip("only for 3.9+", allow_module_level=True)  # type: ignore[var-annotated]

QUERY_ARGUMENTS = "query-arguments"
JSON_BODY_ARGUMENTS = "json-body-arguments"


def setup_argument_handler(
    argument_source: str,
    args: Union[dict[str, Any], list[tuple[Any, Any]]],
) -> ArgumentHandler:
    """Load data and return `ArgumentHandler` instance."""
    if argument_source == QUERY_ARGUMENTS:
        req = requests.Request(url="https://foo.aq/all", params=args).prepare()
        print("\n request:")
        print(req.url)
        print()
        rest_handler = RestHandler(
            application=Mock(),
            request=httputil.HTTPServerRequest(
                uri=req.url,
                headers=httputil.HTTPHeaders(req.headers),
            ),
        )
        return ArgumentHandler(ArgumentSource.QUERY_ARGUMENTS, rest_handler)

    elif argument_source == JSON_BODY_ARGUMENTS:
        req = requests.Request(url="https://foo.aq/all", json=args).prepare()
        print("\n request:")
        print(req.url)
        print(req.body)
        print(req.headers)
        print()
        rest_handler = RestHandler(
            application=Mock(),
            request=httputil.HTTPServerRequest(
                uri=req.url,
                body=cast(bytes, req.body),
                headers=httputil.HTTPHeaders(req.headers),
            ),
        )
        rest_handler.request._parse_body()  # normally called when request REST method starts
        print("\n rest-handler:")
        print(rest_handler.request.body)
        print(json.loads(rest_handler.request.body))
        print()
        return ArgumentHandler(ArgumentSource.JSON_BODY_ARGUMENTS, rest_handler)

    else:
        raise ValueError(f"Invalid argument_source: {argument_source}")


########################################################################################


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_100__defaults(argument_source: str) -> None:
    """Test `argument_source` arguments with default."""
    default: Any

    for default in [None, "a-string", 100, 50.5]:
        print(default)
        arghand = setup_argument_handler(argument_source, {})
        arghand.add_argument("myarg", default=default)
        args = arghand.parse_args()
        assert default == args.myarg

    # w/ typing
    for default in ["a-string", 100, 50.5]:
        print(default)
        arghand = setup_argument_handler(argument_source, {})
        arghand.add_argument("myarg", default=default, type=type(default))
        args = arghand.parse_args()
        assert default == args.myarg


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_110__no_default_no_typing(argument_source: str) -> None:
    """Test `argument_source` arguments."""
    args: dict[str, Any] = {
        "foo": "-10",
        "bar": "True",
        "bat": "2.5",
        "baz": "Toyota Corolla",
        "boo": "False",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "dicto": {"abc": 123},
                "listo": [1, 2, 3],
                "falso": False,
                "truo": True,
                "compoundo": [{"apple": True}, {"banana": 951}, {"cucumber": False}],
            }
        )

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, args)

    for arg, _ in args.items():
        print()
        print(arg)
        arghand.add_argument(arg)
    outargs = arghand.parse_args()

    # grab each
    for arg, val in args.items():
        print()
        print(arg)
        print(val)
        assert val == getattr(outargs, arg.replace("-", "_"))


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_111__no_default_with_typing(argument_source: str) -> None:
    """Test `request.arguments` arguments."""

    def custom_type(_: Any) -> Any:
        return "this could be anything but it's just a string"

    args: dict[str, Any] = {
        "foo": ("-10", int),
        #
        "bar": ("True", bool),
        "boo": ("False", bool),
        "ghi": ("no", bool),
        "jkl": ("1", bool),
        "mno": ("0", bool),
        #
        "bat": ("2.5", float),
        "abc": ("2", float),
        #
        "baz": ("Toyota Corolla", str),
        "def": ("1000", str),
        #
        "zzz": ("yup", custom_type),
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "dicto": ({"abc": 123}, dict),
                "dicto-as-list": ({"def": 456}, list),
                "listo": ([1, 2, 3], list),
                "falso": (False, bool),
                "truo": (True, bool),
                "compoundo": (
                    [{"apple": True}, {"banana": 951}, {"cucumber": False}],
                    list,
                ),
                "sure": (
                    [{"apple": True}, {"banana": 951}, {"cucumber": False}],
                    custom_type,
                ),
            }
        )

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source, {arg: val for arg, (val, _) in args.items()}
    )
    for arg, (_, typ) in args.items():
        print()
        print(arg)
        arghand.add_argument(arg, type=typ)
    outargs = arghand.parse_args()

    # grab each
    for arg, (val, typ) in args.items():
        print()
        print(arg)
        print(val)
        print(typ)
        if typ == bool and isinstance(val, str):
            assert getattr(outargs, arg.replace("-", "_")) == strtobool(val)
        else:
            assert getattr(outargs, arg.replace("-", "_")) == typ(val)


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_112__no_default_with_typing__error(argument_source: str) -> None:
    """Test `argument_source` arguments."""

    def custom_type(val: Any) -> Any:
        raise ValueError("x")

    args: dict[str, Any] = {
        "foo": ("hank", int),
        "bat": ("2.5", int),
        "baz": ("9e-33", int),
        #
        "bar": ("idk", bool),
        "boo": ("2", bool),
        #
        "abc": ("222.111.333", float),
        #
        "zzz": ("yup", custom_type),
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "dicto": ({"abc": 123}, int),
                "listo": ([1, 2, 3], int),
                "compoundo": (
                    [{"apple": True}, {"banana": 951}, {"cucumber": False}],
                    object,
                ),
                "nope": (
                    [{"apple": True}, {"banana": 951}, {"cucumber": False}],
                    custom_type,
                ),
            }
        )

    for arg, (val, typ) in args.items():
        print()
        print(arg)
        print(val)
        print(typ)
        arghand = setup_argument_handler(argument_source, {arg: val})
        arghand.add_argument(arg, type=typ)
        with pytest.raises(tornado.web.HTTPError) as e:
            arghand.parse_args()
        assert str(e.value) == f"HTTP 400: argument {arg}: invalid type"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_120__missing_argument(argument_source: str) -> None:
    """Test `argument_source` arguments error case."""

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, {})
    arghand.add_argument("reqd")

    # Missing Required Argument
    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()
    assert str(e.value) == "HTTP 400: the following arguments are required: reqd"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_121__missing_argument(argument_source: str) -> None:
    """Test `argument_source` arguments error case."""

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, {"foo": "val"})
    arghand.add_argument("reqd")
    arghand.add_argument("foo")
    arghand.add_argument("bar")

    # Missing Required Argument
    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()
    assert str(e.value) == "HTTP 400: the following arguments are required: reqd, bar"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_130__duplicates(argument_source: str) -> None:
    """Test `argument_source` arguments with duplicate keys."""

    args = [
        ("foo", "22"),
        ("foo", "44"),
        ("foo", "66"),
        ("foo", "88"),
        ("bar", "abc"),
        ("baz", "hello world!"),
        ("baz", "hello mars!"),
    ]
    if argument_source == JSON_BODY_ARGUMENTS:
        pass  # no special data since duplicates not allowed for json

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, args)
    arghand.add_argument("foo", type=int, nargs="*")
    arghand.add_argument("baz", type=str, nargs="*")
    arghand.add_argument("bar")

    # duplicates not allowed for json
    if argument_source == JSON_BODY_ARGUMENTS:
        with pytest.raises(tornado.web.HTTPError) as e:
            arghand.parse_args()
        assert str(e.value) == "HTTP 400: JSON-encoded requests body must be a 'dict'"
    # but they are for query args!
    else:
        outargs = arghand.parse_args()
        print(outargs)
        # grab each
        assert outargs.foo == [22, 44, 66, 88]
        assert outargs.baz == ["hello world!", "hello mars!"]
        assert outargs.bar == "abc"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_140__extra_argument(argument_source: str) -> None:
    """Test `argument_source` arguments error case."""

    # set up ArgumentHandler
    args: dict[str, Any] = {
        "foo": "val",
        "reqd": "2",
        "xtra": 1,
        "another": True,
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args["another"] = [{"apple": True}, {"banana": 951}, {"cucumber": False}]

    arghand = setup_argument_handler(argument_source, args)
    arghand.add_argument("reqd")
    arghand.add_argument("foo")

    # Missing Required Argument
    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()
    assert str(e.value) == "HTTP 400: unrecognized arguments: xtra, another"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_141__extra_argument_with_duplicates(argument_source: str) -> None:
    """Test `argument_source` arguments error case."""

    # set up ArgumentHandler
    args = [
        ("foo", "val"),
        ("reqd", "2"),
        ("xtra", 1),
        ("another", True),
        ("another", False),
        ("another", "who knows"),
    ]
    if argument_source == JSON_BODY_ARGUMENTS:
        pass  # no special data since duplicates not allowed for json

    arghand = setup_argument_handler(argument_source, args)
    arghand.add_argument("reqd")
    arghand.add_argument("foo")

    # Missing Required Argument...

    # duplicates not allowed for json
    if argument_source == JSON_BODY_ARGUMENTS:
        with pytest.raises(tornado.web.HTTPError) as e:
            arghand.parse_args()
        assert str(e.value) == "HTTP 400: JSON-encoded requests body must be a 'dict'"
    # but they are for query args!
    else:
        with pytest.raises(tornado.web.HTTPError) as e:
            arghand.parse_args()
        assert str(e.value) == "HTTP 400: unrecognized arguments: xtra, another"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_200__argparse_dest(argument_source: str) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: dict[str, Any] = {
        "old_name": "-10",
        "bar": "True",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "dicto": {"abc": 123},
                "listo": [1, 2, 3],
            }
        )

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, args)

    for arg, _ in args.items():
        print()
        print(arg)
        if arg == "old_name":
            arghand.add_argument(arg, dest="new_name")
        else:
            arghand.add_argument(arg)
    outargs = arghand.parse_args()

    # grab each
    for arg, val in args.items():
        print()
        print(arg)
        print(val)
        if arg == "old_name":
            assert val == getattr(outargs, "new_name")
        else:
            assert val == getattr(outargs, arg.replace("-", "_"))


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_210__argparse_choices(argument_source: str) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: dict[str, Any] = {
        "pick_it": "paper",
        "bar": "True",
    }
    choices: list = ["rock", "paper", "scissors"]
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "pick_it": {"abc": 123},
                "listo": [1, 2, 3],
            }
        )
        choices = [{"abc": 123}, {"def": 456}, {"ghi": 789}]

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, args)

    for arg, _ in args.items():
        print()
        print(arg)
        if arg == "pick_it":
            arghand.add_argument(arg, choices=choices)
        else:
            arghand.add_argument(arg)
    outargs = arghand.parse_args()

    # grab each
    for arg, val in args.items():
        print()
        print(arg)
        print(val)
        assert val == getattr(outargs, arg.replace("-", "_"))


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_211__argparse_choices__error(argument_source: str) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: dict[str, Any] = {
        "pick_it": "paper",
        "bar": "True",
    }
    choices: list[Any] = ["rock", "paper", "scissors"]
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "pick_it": {"abc": 123},
                "listo": [1, 2, 3],
            }
        )
        choices = [{"abc": 123}, {"def": 456}, {"ghi": 789}]

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source,
        {k: v if k != "pick_it" else "hank" for k, v in args.items()},
    )

    for arg, _ in args.items():
        print()
        print(arg)
        if arg == "pick_it":
            arghand.add_argument(arg, choices=choices)
        else:
            arghand.add_argument(arg)

    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()

    # For Python 3.13 and later: join stringified representations directly
    expected_message1 = (
        f"HTTP 400: argument pick_it: invalid choice: 'hank' "
        f"(choose from {', '.join(str(c) for c in choices)})"
    )
    # For Python 3.9–3.12: join repr representations directly
    expected_message2 = (
        f"HTTP 400: argument pick_it: invalid choice: 'hank' "
        f"(choose from {', '.join(repr(c) for c in choices)})"
    )
    assert (
        str(e.value) == expected_message1 or str(e.value) == expected_message2
    ), "Error does not match expected values"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_220__argparse_nargs(argument_source: str) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    test_130__duplicates(argument_source)


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
@pytest.mark.parametrize(
    "exc",
    [TypeError, ValueError, argparse.ArgumentError],
)
def test_230__argparse_custom_validation__supported_builtins__error(
    argument_source: str, exc: Exception
) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: dict[str, Any] = {
        "foo": "True",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args = {
            "foo": [1, 2, 3],
        }

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source,
        args,
    )

    def _error_it(_: Any, exc: Exception):
        raise exc

    for arg, _ in args.items():
        print()
        print(arg)
        arghand.add_argument(
            arg,
            type=lambda x: _error_it(
                x,
                exc("it's a bad value but you won't see this message anyway..."),  # type: ignore
            ),
        )

    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()

    assert str(e.value) == "HTTP 400: argument foo: invalid type"


class MyError(Exception):
    """Used below."""


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
@pytest.mark.parametrize(
    "exc",
    [RuntimeError, IndexError, MyError],
)
def test_232__argparse_custom_validation__unsupported_errors__error(
    argument_source: str, exc: Exception
) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: dict[str, Any] = {
        "foo": "True",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args = {
            "foo": [1, 2, 3],
        }

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source,
        args,
    )

    def _error_it(_: Any, exc: Exception):
        raise exc

    for arg, _ in args.items():
        print()
        print(arg)
        arghand.add_argument(
            arg,
            type=lambda x: _error_it(
                x,
                exc("something went wrong but not in an unexpected way, not-validation"),  # type: ignore
            ),
        )

    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()

    assert re.fullmatch(
        r"HTTP 400: Unknown argument-handling error \(\d+\.\d+\)", str(e.value)
    )


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_234__argparse_custom_validation__argumenttypeerror__error(
    argument_source: str,
) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: dict[str, Any] = {
        "foo": "True",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args = {
            "foo": [1, 2, 3],
        }

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source,
        args,
    )

    def _error_it(_: Any):
        raise argparse.ArgumentTypeError("it's a bad value and you *will* see this!")

    for arg, _ in args.items():
        print()
        print(arg)
        arghand.add_argument(
            arg,
            type=_error_it,
        )

    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()

    assert (
        str(e.value)
        == "HTTP 400: argument foo: it's a bad value and you *will* see this!"
    )


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
@pytest.mark.parametrize(
    "exc",
    [  # all the exceptions!
        TypeError,
        ValueError,
        argparse.ArgumentError,
        argparse.ArgumentTypeError,
        RuntimeError,
        IndexError,
        MyError,
    ],
)
def test_236__argparse_custom_validation__validator_no_param__error(
    argument_source: str,
    exc: Exception,
) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: dict[str, Any] = {
        "foo": "True",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args = {
            "foo": [1, 2, 3],
        }

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source,
        args,
    )

    def _error_it__no_param():
        raise exc("it's a bad value and you *will* see this!")  # type: ignore

    for arg, _ in args.items():
        print()
        print(arg)
        arghand.add_argument(
            arg,
            type=_error_it__no_param,
            # NOTE: ^^^ because this takes no arguments (isn't a func/lambda),
            #       argparse treats it like any other error. why? idk :/
        )

    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()

    assert str(e.value) == "HTTP 400: argument foo: invalid type"



import base64
import hashlib
import secrets
from typing import MutableMapping

from cachetools import TTLCache


class PKCEMixin:
    _pkcs_challenges: MutableMapping[str, str] = TTLCache(maxsize=10000, ttl=3600)

    @classmethod
    def create_pkce_challenge(cls) -> str:
        code_verifier = secrets.token_urlsafe(64)
        code_challenge = base64.urlsafe_b64encode(hashlib.sha256(code_verifier.encode('utf-8')).digest()).decode('utf-8').split('=')[0]
        cls._pkcs_challenges[code_challenge] = code_verifier
        return code_challenge

    @classmethod
    def get_pkce_verifier(cls, challenge: str) -> str:
        if challenge in cls._pkcs_challenges:
            return cls._pkcs_challenges[challenge]
        raise KeyError('invalid pkce challenge')



"""A simple REST json client using `requests`_ for the http connection.

.. _requests: http://docs.python-requests.org

The REST protocol is built on http(s), with the body containing
a json-encoded dictionary as necessary.
"""

# fmt:quotes-ok

import asyncio
import dataclasses as dc
import logging
import math
import os
import time
from typing import Any, Callable, Generator, Optional, Union

import jwt
import requests
import urllib3.util

from .. import telemetry as wtt
from ..utils.json_util import JSONType, json_decode
from .session import AsyncSession, Session

MAX_RETRIES = 30


def _to_str(s: Union[str, bytes]) -> str:
    if isinstance(s, bytes):
        return s.decode('utf-8')
    return s


@dc.dataclass
class CalcRetryFromBackoffMax:
    """An indicator to auto-calculate the # of retries using a backoff_max.

    `backoff_max` is the maximum allowed backoff time for the final
    retry
    """

    backoff_max: float

    def __post_init__(self) -> None:
        if self.backoff_max > urllib3.util.retry.Retry.DEFAULT_BACKOFF_MAX:
            raise ValueError(
                f"CalcRetryFromBackoffMax.backoff_max ({self.backoff_max})"
                f" cannot be greater than: {urllib3.util.retry.Retry.DEFAULT_BACKOFF_MAX=}"
            )

    def calculate_retries(self, backoff_factor: float) -> int:
        """Calculate the number of retries."""
        #                        backoff_last     =  backoff_factor * 2^(retries-1)
        #       backoff_last / backoff_factor     =  2^(retries-1)
        # lg( backoff_last / backoff_factor )     =  retries - 1
        # lg( backoff_last / backoff_factor ) + 1 =  retries - 1
        return max(
            0,
            int(math.log2(self.backoff_max / backoff_factor) + 1),
        )


@dc.dataclass
class CalcRetryFromWaittimeMax:
    """An indicator to auto-calculate the # of retries from total waittime.

    `waittime_max` includes each attempt's timeout and each backoff time
    """

    waittime_max: float

    def calculate_retries(self, timeout: float, backoff_factor: float) -> int:
        """Calculate the number of retries (max returned -> MAX_RETRIES+1)."""
        if self.waittime_max < timeout:
            raise ValueError(
                f"waittime_max ({self.waittime_max}) cannot be less than timeout ({timeout})"
            )

        # the first backoff is always 0 sec, factor applies after 2nd attempt
        #    T  +  0  +  sum{n=1,retries-1}(T + min[MAX, 2^n * B] )  +  T
        # sum has no closed form due to `min` function
        retries = 0
        for candidate in range(0, MAX_RETRIES + 2):  # last val is MAX_RETRIES+1
            total = (
                timeout
                + 0
                + sum(
                    timeout
                    + min(
                        urllib3.util.retry.Retry.DEFAULT_BACKOFF_MAX,
                        2**n * backoff_factor,
                    )
                    for n in range(1, candidate)  # 1 to retries-1 (inclusive)
                )
                + timeout
            )
            # print(candidate)
            # print(total)
            if self.waittime_max > total:
                retries = candidate  # gets overwritten each iteration
            else:
                break
        return retries


def _log_retries_values(
    retries: int, timeout: float, backoff_factor: float, logger: logging.Logger
) -> None:
    logger.debug(f"using {retries=} {timeout=} {backoff_factor=}")
    if retries:
        retries_schema = ' '.join(
            # first backoff is 0
            [f'<0.0s> {timeout}s']
            +
            # calculate increasing backoff values
            [
                f'<{min(urllib3.util.retry.Retry.DEFAULT_BACKOFF_MAX,backoff_factor * 2**i)}s> {timeout}s'
                for i in range(1, retries)
            ]
        )
        logger.debug(
            f"retry scheme (TIMEOUT [<BACKOFF> TIMEOUT ...]): "
            f"{timeout}s {retries_schema}"
        )


class RestClient:
    """A REST client with token handling.

    Args:
        address (str):
            base address of REST API
        token (str):
            (optional) access token, or a function generating an access token
        timeout (int):
            (optional) request timeout (default: 60s)
        retries (int | CalcRetryFromBackoffMax | CalcRetryFromWaittimeMax):
            (optional) number of retries to attempt (default: 10)
            alternatively, pass in a `CalcRetryFromBackoffMax` or
            `CalcRetryFromWaittimeMax` instance to have this auto-calculated
        backoff_factor (float):
            (optional) backoff factor to apply between attempts after the second try --
            sleep for `{backoff factor} * (2 ** ({number of previous retries}))` seconds
            "For example, if the backoff_factor is 0.1, then Retry.sleep() will sleep for [0.0s, 0.2s, 0.4s, 0.8s, ...] between retries."
            see: https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#module-urllib3.util.retry
        username (str):
            (optional) auth-basic username
        password (str):
            (optional) auth-basic password
        logger (logging.Logger):
            (optional) supply a logger to use
    """

    def __init__(
        self,
        address: str,
        token: Optional[Union[str, bytes, Callable[[], Union[str, bytes]]]] = None,
        timeout: float = 60.0,
        retries: Union[int, CalcRetryFromBackoffMax, CalcRetryFromWaittimeMax] = 10,
        backoff_factor: float = 0.3,
        logger: Optional[logging.Logger] = None,
        **kwargs: Any,
    ) -> None:
        self.address = address
        self.kwargs = kwargs
        self.logger = logger if logger else logging.getLogger('RestClient')

        self.timeout = float(timeout)
        if self.timeout < 0.0:
            raise ValueError(f"timeout must be positive: {self.timeout}")

        self.backoff_factor = float(backoff_factor)
        if self.backoff_factor < 0.0:
            raise ValueError(f"backoff_factor must be positive: {self.backoff_factor}")

        # get numerical retries value
        if isinstance(retries, CalcRetryFromBackoffMax):
            self.retries = retries.calculate_retries(self.backoff_factor)
        elif isinstance(retries, CalcRetryFromWaittimeMax):
            self.retries = retries.calculate_retries(self.timeout, self.backoff_factor)
        elif isinstance(retries, int) and retries >= 0:
            self.retries = retries
        else:
            raise ValueError(
                f"Cannot set and/or auto-calculate # of retries: {retries}"
            )
        # + validate
        if self.retries > MAX_RETRIES:
            raise ValueError(f"Cannot set # of retries above {MAX_RETRIES}")
        _log_retries_values(
            self.retries, self.timeout, self.backoff_factor, self.logger
        )

        # token handling
        self._token_expire_delay_offset = 5
        self.access_token: Optional[Union[str, bytes]] = None
        self.token_func: Optional[Callable[[], Union[str, bytes]]] = None
        if token:
            if isinstance(token, (str, bytes)):
                self.access_token = token
            elif callable(token):
                self.token_func = token

        self.session = self.open()  # start session

    def open(self, sync: bool = False) -> requests.Session:
        """Open the http session."""
        self.logger.debug('establish REST http session')
        if sync:
            self.session = Session(
                self.retries,
                backoff_factor=self.backoff_factor,
            )
        else:
            self.session = AsyncSession(
                self.retries,
                backoff_factor=self.backoff_factor,
            )
        self.session.headers = {  # type: ignore[assignment]
            'Content-Type': 'application/json',
        }
        if 'username' in self.kwargs and 'password' in self.kwargs:
            self.session.auth = (self.kwargs['username'], self.kwargs['password'])
        if 'sslcert' in self.kwargs:
            if 'sslkey' in self.kwargs:
                self.session.cert = (self.kwargs['sslcert'], self.kwargs['sslkey'])
            else:
                self.session.cert = self.kwargs['sslcert']
        if 'cacert' in self.kwargs:
            self.session.verify = self.kwargs['cacert']

        return self.session

    def close(self) -> None:
        """Close the http session."""
        self.logger.info('close REST http session')
        if self.session:
            self.session.close()

    def _get_token(self) -> None:
        if self.access_token:
            # check if expired
            try:
                # NOTE: PyJWT mis-type-hinted arg #1 as a str, but byte is also fine
                # https://github.com/jpadilla/pyjwt/pull/605#issuecomment-772082918
                data = jwt.decode(
                    self.access_token,  # type: ignore[arg-type]
                    algorithms=['RS256', 'RS512'],
                    options={"verify_signature": False},
                )
                # account for an X second delay over the wire, so expire sooner
                if data['exp'] < time.time() + self._token_expire_delay_offset:
                    raise Exception()
                return
            except Exception:
                self.access_token = None
                self.logger.debug('token expired')

        try:
            self.access_token = self.token_func()  # type: ignore[misc]
        except Exception:
            self.logger.warning('acquiring access token failed')
            raise

    def _prepare(
        self,
        method: str,
        path: str,
        args: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, str]] = None,
    ) -> tuple[str, dict[str, Any]]:
        """Internal method for preparing requests."""
        if not args:
            args = {}

        # auto-inject the current span's info into the HTTP headers
        wtt.inject_span_carrier_if_recording(self.session.headers)  # type: ignore[arg-type]

        if path.startswith('/'):
            path = path[1:]
        url = os.path.join(self.address, path)

        kwargs: dict[str, Any] = {'timeout': self.timeout}

        if method in ('GET', 'HEAD'):
            # args should be urlencoded
            kwargs['params'] = args
        else:
            kwargs['json'] = args

        if self.token_func:
            self._get_token()

        if not headers:
            headers = {}

        if self.access_token:
            headers['Authorization'] = 'Bearer ' + _to_str(self.access_token)

        if headers:
            kwargs['headers'] = headers

        return (url, kwargs)

    def _decode(self, content: Union[str, bytes, bytearray]) -> JSONType:
        """Internal method for translating response from json."""
        if not content:
            self.logger.info('request returned empty string')
            return None
        try:
            return json_decode(content)
        except Exception:
            self.logger.info('json data: %r', content)
            raise

    @wtt.spanned(
        span_namer=wtt.SpanNamer(use_this_arg='method'),
        these=['method', 'path', 'self.address'],
        kind=wtt.SpanKind.CLIENT,
    )
    async def request(
        self,
        method: str,
        path: str,
        args: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, str]] = None,
    ) -> JSONType:
        """Send request to REST Server.

        Async request - use with coroutines.

        Args:
            method (str): the http method
            path (str): the url path on the server
            args (dict): any arguments to pass
            headers (dict): any headers to pass to the request

        Returns:
            dict: json dict or raw string
        """
        url, kwargs = self._prepare(method, path, args, headers)
        try:
            # session: AsyncSession; So, self.session.request() -> Future
            r: requests.Response = await asyncio.wrap_future(self.session.request(method, url, **kwargs))  # type: ignore[arg-type]
            r.raise_for_status()
            return self._decode(r.content)
        except requests.exceptions.HTTPError as e:
            if method == 'DELETE' and e.response.status_code == 404:
                raise  # skip the logging for an expected error
            self.logger.info('bad request: %s %s %r', method, path, args, exc_info=True)
            raise

    @wtt.spanned(
        span_namer=wtt.SpanNamer(use_this_arg='method'),
        these=['method', 'path', 'self.address'],
        kind=wtt.SpanKind.CLIENT,
    )
    def request_seq(
        self,
        method: str,
        path: str,
        args: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, str]] = None,
    ) -> JSONType:
        """Send request to REST Server.

        Sequential version of `request`.

        Args:
            method (str): the http method
            path (str): the url path on the server
            args (dict): any arguments to pass
            headers (dict): any headers to pass to the request

        Returns:
            dict: json dict or raw string
        """
        s = self.session
        try:
            self.open(sync=True)
            url, kwargs = self._prepare(method, path, args, headers)
            r: requests.Response = self.session.request(method, url, **kwargs)  # type: ignore
            r.raise_for_status()
            return self._decode(r.content)
        finally:
            self.session = s

    @wtt.spanned(
        span_namer=wtt.SpanNamer(use_this_arg='method'),
        these=['method', 'path', 'self.address'],
        kind=wtt.SpanKind.CLIENT,
    )
    def request_stream(
        self,
        method: str,
        path: str,
        args: Optional[dict[str, Any]] = None,
        headers: Optional[dict[str, str]] = None,
        chunk_size: Optional[int] = 8096,
    ) -> Generator[JSONType, None, None]:
        """Send request to REST Server, and stream results back.

        `chunk_size=None` will read data as it arrives
        in whatever size the chunks are received. `chunk_size`<`1`
        will be treated as `chunk_size=None`

        Args:
            method (str): the http method
            path (str): the url path on the server
            args (dict): any arguments to pass
            headers (dict): any headers to pass to the request
            chunk_size (int): chunk size (see above)

        Returns:
            dict: json dict or raw string
        """
        if chunk_size is not None and chunk_size < 1:
            chunk_size = None

        s = self.session
        try:
            self.open(sync=True)
            url, kwargs = self._prepare(method, path, args, headers)
            resp: requests.Response = self.session.request(method, url, stream=True, **kwargs)  # type: ignore
            resp.raise_for_status()
            for line in resp.iter_lines(chunk_size=chunk_size, delimiter=b'\n'):
                decoded = self._decode(line.strip())
                if decoded:  # skip `None`
                    yield decoded
        finally:
            self.session = s



"""Example client code.

Add some fruit to the server, then read them back.
"""

import asyncio
import logging

from rest_tools.client import RestClient  # noqa: E402 # pylint: disable=C0413,E0401
from rest_tools.utils import Auth  # noqa: E402 # pylint: disable=C0413,E0401
from wipac_telemetry import tracing_tools  # type: ignore[import]


@tracing_tools.spanned()
async def main() -> None:
    """Establish REST connection, make some requests, handle responses."""
    admin_token = Auth("secret").create_token("foo", payload={"role": "admin"})
    user_token = Auth("secret").create_token("foo", payload={"role": "user"})

    rc = RestClient("http://localhost:8080/api", token=admin_token)
    rc_user = RestClient("http://localhost:8080/api", token=user_token)

    # populate
    await rc.request("POST", "/fruits", {"name": "apple"})
    await rc.request("POST", "/fruits", {"name": "banana"})

    # # query
    ret = await rc_user.request("GET", "/fruits")
    assert ret["fruits"] == {"apple": {"name": "apple"}, "banana": {"name": "banana"}}


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    asyncio.run(main())



"""Integration test fixtures."""

import socket

import pytest


@pytest.fixture
def port() -> int:
    """Get an ephemeral port number."""
    # unix.stackexchange.com/a/132524
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    addr = s.getsockname()
    ephemeral_port = addr[1]
    s.close()
    return ephemeral_port



"""Test script for auth."""

# fmt:off
# pylint: skip-file

import time

import jwt
import pytest

# local imports
from rest_tools.utils import auth

from .fixtures import (  # noqa: F401 # gen_keys_bytes uses gen_keys
    gen_keys,
    gen_keys_bytes,
)


def test_auth_create_token():
    a = auth.Auth('secret')
    now = time.time()
    tok = a.create_token('subj', expiration=20, payload={'type':'foo'})

    data = jwt.decode(tok, 'secret', algorithms=['HS512'])
    assert data['sub'] == 'subj'
    assert data['type'] == 'foo'
    assert data['exp'] < now+21


def test_auth_validate():
    a = auth.Auth('secret')
    tok = a.create_token('subj', expiration=20)
    data = a.validate(tok)
    assert data['sub'] == 'subj'


def test_auth_validate_leeway():
    a = auth.Auth('secret', leeway=0)
    tok = a.create_token('subj', expiration=-1)
    with pytest.raises(jwt.exceptions.ExpiredSignatureError):
        a.validate(tok)

    a = auth.Auth('secret', leeway=10)
    tok = a.create_token('subj', expiration=-1)
    data = a.validate(tok)
    assert data['sub'] == 'subj'


def test_auth_validate_exp_int():
    a = auth.Auth('secret', integer_times=False)
    tok = a.create_token('subj', expiration=125)
    data = a.validate(tok)
    assert isinstance(data['exp'], float)

    a = auth.Auth('secret', integer_times=True)
    tok = a.create_token('subj', expiration=125)
    data = a.validate(tok)
    assert isinstance(data['exp'], int)


def test_auth_validate_aud():
    a = auth.Auth('secret', audience=['bar'])
    tok = a.create_token('subj', expiration=20, payload={'aud': 'foo'})
    with pytest.raises(jwt.exceptions.InvalidAudienceError):
        a.validate(tok)

    a.validate(tok, audience='foo')


def test_auth_validate_aud_none():
    a = auth.Auth('secret', audience=None)
    tok = a.create_token('subj', expiration=20)
    a.validate(tok)


def test_auth_validate_iss():
    a = auth.Auth('secret', issuer='foo')
    tok = a.create_token('subj', expiration=20)
    data = a.validate(tok)
    assert data['iss'] == 'foo'

    with pytest.raises(jwt.exceptions.InvalidIssuerError):
        a._validate(tok, 'secret', issuers=['bar'])


def test_auth_validate_iss_none():
    a = auth.Auth('secret', issuer='foo')
    tok = a.create_token('subj', expiration=20)
    data = a._validate(tok, 'secret')
    assert data['iss'] == 'foo'


def test_auth_rsa(gen_keys_bytes):  # noqa: F811
    a = auth.Auth(gen_keys_bytes[0], pub_secret=gen_keys_bytes[1], algorithm='RS256')
    tok = a.create_token('subj', expiration=20)
    data = jwt.decode(tok, gen_keys_bytes[1], algorithms=['RS256'])
    assert data['sub'] == 'subj'

    data = a.validate(tok)
    assert data['sub'] == 'subj'


def test_auth_rsa_aud_iss(gen_keys_bytes):  # noqa: F811
    a = auth.Auth(gen_keys_bytes[0], pub_secret=gen_keys_bytes[1], issuer='foo', audience=['bar'], issuers=['foo'], algorithm='RS256')
    tok = a.create_token('subj', expiration=20, payload={'aud': 'bar'})
    a.validate(tok)



"""Test route handlers for the OpenAPI spec validation."""

import re
from typing import AsyncIterator, Callable

import openapi_core
import pytest
import pytest_asyncio
import requests
from jsonschema_path import SchemaPath

from rest_tools.client import RestClient
from rest_tools.server import RestServer, RestHandler, validate_request


@pytest_asyncio.fixture
async def server(port: int) -> AsyncIterator[Callable[[], RestClient]]:
    """Start up REST server and attach handlers."""
    rs = RestServer(debug=True)

    class FooNoArgsHandler(RestHandler):
        ROUTE = "/foo/no-args"

        # NOTE: there is no way in openapi v3.1 to explicitly forbid extra query params
        # @validate_request(OPENAPI_SPEC)
        # async def get(self) -> None:
        #     self.write({"message": "hello world"})

        @validate_request(OPENAPI_SPEC)
        async def post(self) -> None:
            self.write({"message": "hello world"})

    class FooURLParamsHandler(RestHandler):
        ROUTE = r"/foo/params/(?P<the_id>\w+)/(?P<the_name>\w+)$"

        @validate_request(OPENAPI_SPEC)
        async def get(self, the_id: int, the_name: str) -> None:
            self.write({"message": f"got {the_id} {the_name}"})

        @validate_request(OPENAPI_SPEC)
        async def post(self, the_id: int, the_name: str) -> None:
            self.write({"message": f"posted {the_id} {the_name}"})

    class FooArgsHandler(RestHandler):
        ROUTE = "/foo/args"

        @validate_request(OPENAPI_SPEC)
        async def get(self) -> None:
            # args embedded in url's query string
            self.write({"message": f"hello {self.get_argument('rank')}"})

        @validate_request(OPENAPI_SPEC)
        async def post(self) -> None:
            # args in JSON
            self.write({"message": f"hey {self.get_argument('rank')}"})

    class FooURLParamsAndArgsHandler(RestHandler):
        ROUTE = r"/foo/params-and-args/(?P<the_id>\w+)/(?P<the_name>\w+)$"

        @validate_request(OPENAPI_SPEC)
        async def get(self, the_id: str, the_name: str) -> None:
            # args embedded in url's query
            self.write(
                {"message": f"got {the_id} {the_name} -- {self.get_argument('why')}"}
            )

        @validate_request(OPENAPI_SPEC)
        async def post(self, the_id: str, the_name: str) -> None:
            # args in JSON
            self.write(
                {"message": f"posted {the_id} {the_name} -- {self.get_argument('why')}"}
            )

    rs.add_route(FooNoArgsHandler.ROUTE, FooNoArgsHandler)
    rs.add_route(FooURLParamsHandler.ROUTE, FooURLParamsHandler)
    rs.add_route(FooArgsHandler.ROUTE, FooArgsHandler)
    rs.add_route(FooURLParamsAndArgsHandler.ROUTE, FooURLParamsAndArgsHandler)
    rs.startup(address="localhost", port=port)

    def client() -> RestClient:
        return RestClient(f"http://localhost:{port}", retries=0)

    try:
        yield client
    finally:
        await rs.stop()


OPENAPI_SPEC = openapi_core.OpenAPI(
    SchemaPath.from_dict(
        {
            "openapi": "3.1.0",
            "info": {"title": "Foo API", "version": "1.0.0"},
            "components": {
                "parameters": {
                    "TheNameParam": {
                        "name": "the_name",
                        "in": "path",
                        "required": True,
                        "schema": {"type": "string"},
                    },
                },
                "schemas": {
                    "InfoObject": {
                        "type": "object",
                        "properties": {
                            "alias": {"type": "string"},
                            "rank": {"type": "integer"},
                        },
                    },
                },
            },
            "paths": {
                "/foo/no-args": {
                    "post": {
                        "requestBody": {
                            "content": {
                                "application/json": {
                                    "schema": {  # no args allowed
                                        "type": "object",
                                        "properties": {},
                                        "required": [],
                                        "additionalProperties": False,
                                    }
                                }
                            }
                        },
                        # "responses": {},  # we're not validating/testing this, so don't bother
                    },
                },
                "/foo/params/{the_id}/{the_name}": {
                    "parameters": [
                        {
                            "name": "the_id",
                            "in": "path",
                            "required": True,
                            "schema": {"type": "integer"},
                        },
                        {"$ref": "#/components/parameters/TheNameParam"},
                    ],
                    "get": {
                        # "responses": {},  # we're not validating/testing this, so don't bother
                    },
                    "post": {
                        # "responses": {},  # we're not validating/testing this, so don't bother
                    },
                },
                "/foo/args": {
                    "get": {
                        "parameters": [
                            {
                                "name": "rank",
                                "in": "query",
                                "required": True,
                                "schema": {"type": "integer"},
                            }
                        ],
                        # "responses": {},  # we're not validating/testing this, so don't bother
                    },
                    "post": {
                        "requestBody": {
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                        "properties": {
                                            "rank": {
                                                "$ref": "#/components/schemas/InfoObject/properties/rank"
                                            },
                                        },
                                        "required": ["rank"],
                                        "additionalProperties": False,
                                    }
                                }
                            }
                        },
                        # "responses": {},  # we're not validating/testing this, so don't bother
                    },
                },
                "/foo/params-and-args/{the_id}/{the_name}": {
                    "parameters": [
                        {
                            "name": "the_id",
                            "in": "path",
                            "required": True,
                            "schema": {"type": "integer"},
                        },
                        {"$ref": "#/components/parameters/TheNameParam"},
                    ],
                    "get": {
                        "parameters": [
                            {
                                "name": "why",
                                "in": "query",
                                "required": True,
                                "schema": {"type": "string"},
                            }
                        ],
                        # "responses": {},  # we're not validating/testing this, so don't bother
                    },
                    "post": {
                        "requestBody": {
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                        "properties": {"why": {"type": "string"}},
                                    }
                                }
                            }
                        },
                        # "responses": {},  # we're not validating/testing this, so don't bother
                    },
                },
            },
        },
    )
)


async def test_000__valid(server: Callable[[], RestClient]) -> None:
    """Test server handler methods with valid data."""
    rc = server()

    # no args allowed
    res = await rc.request("POST", "/foo/no-args")
    assert res == {"message": "hello world"}

    # url params
    res = await rc.request("GET", "/foo/params/123/hank")
    assert res == {"message": "got 123 hank"}
    res = await rc.request("POST", "/foo/params/456/tilly")
    assert res == {"message": "posted 456 tilly"}

    # args
    res = await rc.request("GET", "/foo/args", {"rank": 123})
    assert res == {"message": "hello 123"}
    res = await rc.request("POST", "/foo/args", {"rank": 456})
    assert res == {"message": "hey 456"}

    # args + url params
    res = await rc.request(
        "GET", "/foo/params-and-args/789/book", {"why": "the future"}
    )
    assert res == {"message": "got 789 book -- the future"}
    res = await rc.request("POST", "/foo/params-and-args/248/saru", {"why": "the past"})
    assert res == {"message": "posted 248 saru -- the past"}


async def test_010__invalid(server: Callable[[], RestClient]) -> None:
    """Test server handler methods with invalid data."""
    rc = server()

    #
    # no args allowed
    #

    # POST
    with pytest.raises(
        requests.HTTPError,
        match=re.escape(
            f"Additional properties are not allowed ('have' was unexpected) for url: {rc.address}/foo/no-args"
        ),
    ) as e:
        # extra arg(s)
        await rc.request("POST", "/foo/no-args", {"have": "some args"})
    assert e.value.response.status_code == 400

    #
    # url params
    #

    # GET
    with pytest.raises(
        requests.HTTPError,
        match=re.escape(
            f"Path parameter error: the_id for url: {rc.address}/foo/params/abc/888"
        ),
    ) as e:
        # bad type
        await rc.request("GET", "/foo/params/abc/888")
    assert e.value.response.status_code == 400

    # POST
    with pytest.raises(
        requests.HTTPError,
        match=re.escape(
            f"Path parameter error: the_id for url: {rc.address}/foo/params/xyz/999"
        ),
    ) as e:
        # bad type
        await rc.request("POST", "/foo/params/xyz/999")
    assert e.value.response.status_code == 400

    #
    # args
    #

    # GET
    with pytest.raises(
        requests.HTTPError,
        match=re.escape(
            f"Missing required query parameter: rank for url: {rc.address}/foo/args"
        ),
    ) as e:
        # missing arg(s)
        await rc.request("GET", "/foo/args")
    assert e.value.response.status_code == 400
    # with pytest.raises(requests.HTTPError) as e:
    #     # extra arg(s) -- NOTE: THESE ARE ACTUALLY OK
    #     await rc.request("GET", "/foo/args", {"name": "dwayne", "car": "vroom"})
    with pytest.raises(
        requests.HTTPError,
        match=re.escape(
            f"Query parameter error: rank for url: {rc.address}/foo/args?rank=abc"
        ),
    ) as e:
        # bad type
        await rc.request("GET", "/foo/args", {"rank": "abc"})
    assert e.value.response.status_code == 400

    # POST
    with pytest.raises(
        requests.HTTPError,
        match=re.escape(
            f"'rank' is a required property for url: {rc.address}/foo/args"
        ),
    ) as e:
        # missing arg(s)
        await rc.request("POST", "/foo/args")
    assert e.value.response.status_code == 400
    #
    with pytest.raises(
        requests.HTTPError,
        match=re.escape(
            f"Additional properties are not allowed ('suv' was unexpected) for url: {rc.address}/foo/args"
        ),
    ) as e:
        # extra arg(s)
        await rc.request("POST", "/foo/args", {"rank": 123, "suv": "gas"})
    assert e.value.response.status_code == 400
    #
    with pytest.raises(
        requests.HTTPError,
        match=re.escape(
            f"Request body validation error for url: {rc.address}/foo/args"
        ),
    ) as e:
        # bad type
        await rc.request("POST", "/foo/args", {"rank": "abc"})
    assert e.value.response.status_code == 400

    #
    # args + url params
    #

    # NOTE: not testing the compound cases, since that's exponentially more tests for
    #    little work. By now, we can safely assume that those cases are good since their
    #    components are *independent* (url params and args handling logics are independent)



from unittest.mock import AsyncMock, MagicMock

import pytest
from tornado.web import HTTPError

from rest_tools.server import decorators


async def test_authenticated():
    mock = AsyncMock()
    f = decorators.authenticated(mock)

    self = MagicMock()
    self.current_user = None
    with pytest.raises(HTTPError, match='authentication failed'):
        await f(self)
    mock.assert_not_awaited()

    self.current_user = 'sub'
    self.auth_data = {'sub': 'sub'}
    await f(self)
    mock.assert_awaited()

    await f(self, 1, a='b')
    mock.assert_called_with(self, 1, a='b')


async def test_catch_error():
    @decorators.catch_error
    async def test_passthrough(self):
        raise HTTPError(400, reason='foo')

    self = MagicMock()
    self.send_error = MagicMock()
    with pytest.raises(HTTPError, match='foo'):
        await test_passthrough(self)
    self.send_error.assert_not_called()

    @decorators.catch_error
    async def test_error(self):
        raise Exception('blah')

    await test_error(self)
    self.send_error.assert_called_with(500, reason='Error in MagicMock')


async def test_role_authorization():
    mock = AsyncMock()
    f = decorators.role_authorization(roles=['write'])(mock)

    self = MagicMock()
    self.current_user = None
    with pytest.raises(HTTPError, match='authentication failed'):
        await f(self)
    mock.assert_not_awaited()

    self = MagicMock()
    self.current_user = 'sub'
    self.auth_data = {'sub': 'sub'}
    with pytest.raises(HTTPError, match='authorization failed'):
        await f(self)
    mock.assert_not_awaited()

    self.current_user = 'sub'
    self.auth_data = {'sub': 'sub', 'role': 'read'}
    with pytest.raises(HTTPError, match='authorization failed'):
        await f(self)
    mock.assert_not_awaited()

    self.auth_data = {'sub': 'sub', 'role': 'write'}
    await f(self)
    mock.assert_awaited()

    await f(self, 1, a='b')
    mock.assert_called_with(self, 1, a='b')


async def test_scope_role_auth():
    mock = AsyncMock()
    f = decorators.scope_role_auth(roles=['write'], prefix='test')(mock)

    self = MagicMock()
    self.current_user = None
    with pytest.raises(HTTPError, match='authentication failed'):
        await f(self)
    mock.assert_not_awaited()

    self = MagicMock()
    self.current_user = 'sub'
    self.auth_data = {'sub': 'sub'}
    with pytest.raises(HTTPError, match='authorization failed'):
        await f(self)
    mock.assert_not_awaited()

    self.current_user = 'sub'
    self.auth_data = {'sub': 'sub', 'scope': 'test:read'}
    with pytest.raises(HTTPError, match='authorization failed'):
        await f(self)
    mock.assert_not_awaited()

    self.current_user = 'sub'
    self.auth_data = {'sub': 'sub', 'scope': 'test:read foo:write write'}
    with pytest.raises(HTTPError, match='authorization failed'):
        await f(self)
    mock.assert_not_awaited()

    self.auth_data = {'sub': 'sub', 'scope': 'test:read foo:write test:write'}
    await f(self)
    mock.assert_awaited()

    await f(self, 1, a='b')
    mock.assert_called_with(self, 1, a='b')


async def test_keycloak_role_auth():
    mock = AsyncMock()
    f = decorators.keycloak_role_auth(roles=['write'], prefix='my.roles')(mock)

    self = MagicMock()
    self.current_user = None
    with pytest.raises(HTTPError, match='authentication failed'):
        await f(self)
    mock.assert_not_awaited()

    self = MagicMock()
    self.current_user = 'sub'
    self.auth_data = {'sub': 'sub'}
    with pytest.raises(HTTPError, match='authorization failed'):
        await f(self)
    mock.assert_not_awaited()

    self.current_user = 'sub'
    self.auth_data = {'sub': 'sub', 'my': {'roles': ['read']}}
    with pytest.raises(HTTPError, match='authorization failed'):
        await f(self)
    mock.assert_not_awaited()

    self.auth_data = {'sub': 'sub', 'my': {'roles': ['write']}}
    await f(self)
    mock.assert_awaited()

    await f(self, 1, a='b')
    mock.assert_called_with(self, 1, a='b')


async def test_token_attribute_role_mapping_auth():
    mock = AsyncMock()
    d = decorators.token_attribute_role_mapping_auth(
        role_attrs={'write': ['my.roles=write']},
    )
    f = d(roles=['write'])(mock)

    self = MagicMock()
    self.current_user = None
    with pytest.raises(HTTPError, match='authentication failed'):
        await f(self)
    mock.assert_not_awaited()

    self = MagicMock()
    self.current_user = 'sub'
    self.auth_data = {'sub': 'sub'}
    with pytest.raises(HTTPError, match='authorization failed'):
        await f(self)
    mock.assert_not_awaited()

    self.current_user = 'sub'
    self.auth_data = {'sub': 'sub', 'my': {'roles': ['read']}}
    with pytest.raises(HTTPError, match='authorization failed'):
        await f(self)
    mock.assert_not_awaited()

    self.auth_data = {'sub': 'sub', 'my': {'roles': ['write']}}
    await f(self)
    mock.assert_awaited()

    await f(self, 1, a='b')
    mock.assert_called_with(self, 1, a='b')


async def test_token_attribute_role_mapping_auth_groups():
    mock = AsyncMock()
    d = decorators.token_attribute_role_mapping_auth(
        role_attrs={r'\1': ['my.roles=(write)', 'my.roles=(read)']},
        group_attrs={r'\1': [r'groups=(\w+)']},
    )
    f = d(roles=['write'])(mock)

    self = MagicMock()
    self.auth_data = {'sub': 'sub', 'my': {'roles': ['write']}, 'groups': ['foo', 'bar']}
    await f(self, 1, a='b')
    mock.assert_called_with(self, 1, a='b')
    assert self.auth_groups == ['bar', 'foo']  #: in sorted order



"""
Session storage.

Memory is usually for testing.

Redis is for production, and ideal when mulitple servers are running.
"""

from collections.abc import MutableMapping
from collections import UserDict
import dataclasses as dc
from enum import Enum
import logging
import time
from typing import Iterator, Union


SessionDataTypes = Union[bool, float, int, str]
SessionData = MutableMapping[str, SessionDataTypes]


# Define the structure of the internal session representation
@dc.dataclass(frozen=True)
class SessionEntry:
    data: SessionData
    expiration: float


# Define the session storage type
class SessionStorage(MutableMapping[str, SessionEntry]):
    def close(self):
        ...


class MemorySessionStorage(UserDict[str, SessionEntry], SessionStorage):
    """
    In-memory session storage.

    This is just a dict, and is designed for testing.
    """
    def close(self):
        pass


try:
    import redis
    from redis.cache import CacheConfig
    from redis.commands.search.field import TextField
    from redis.commands.search.index_definition import IndexDefinition, IndexType
    from redis.backoff import ExponentialBackoff
    from redis.retry import Retry
    import redis.exceptions
    redis_available = True
except ImportError:
    redis_available = False
else:
    class RedisSessionStorage(SessionStorage):
        """
        Redis-backed session storage.

        Ideal for production, where multiple servers could be running at the same time.
        """
        def __init__(self):
            retry = Retry(ExponentialBackoff(), 5)
            self._conn = redis.Redis(
                cache_config=CacheConfig(),
                decode_responses=True,
                protocol=3,
                retry=retry,
            )
            self._conn.ping()
            try:
                self._conn.ft('idx:key').info()
            except redis.exceptions.ResponseError as e:
                if 'Unknown Index Name' in str(e):
                    schema = (
                        TextField('key'),
                    )
                    # Create the index
                    self._conn.ft('idx:key').create_index(
                        schema,  # type: ignore
                        definition=IndexDefinition(
                            prefix=['session:'], index_type=IndexType.HASH
                        )
                    )

        def close(self):
            self._conn.close()

        def __getitem__(self, key: str) -> SessionEntry:
            ret: dict = self._conn.hgetall('session:'+key)  # type: ignore
            if not ret:
                raise KeyError()
            exp = float(ret.pop('_expiration'))
            return SessionEntry(data=ret, expiration=exp)

        def __setitem__(self, key: str, value: SessionEntry):
            data = {'_expiration': value.expiration}
            data.update(value.data)  # type: ignore
            # we delete first to clear extra keys not in the new data
            self._conn.delete('session:'+key)
            self._conn.hset('session:'+key, mapping=data)

        def __delitem__(self, key: str):
            self._conn.delete('session:'+key)

        def __iter__(self) -> Iterator:
            for key in self._conn.keys('session:*'):  # type: ignore
                yield key[8:]

        def __len__(self) -> int:
            return self._conn.dbsize()  # type: ignore


class StorageTypes(Enum):
    MEMORY = 'memory'
    REDIS = 'redis'


class Session:
    """
    Session storage with expiration.
    """
    _sessions: SessionStorage

    def __init__(self, storage_type: Union[str, StorageTypes] = 'memory', expiration: float = 1800):
        """
        Initializes the session store.

        Args:
            expiration: The time in seconds after which a session will expire
                        due to inactivity. Defaults to 1800 (30 minutes).
        """
        st = StorageTypes(storage_type)
        if st == StorageTypes.MEMORY:
            self._sessions = MemorySessionStorage()
        elif st == StorageTypes.REDIS:
            if not redis_available:
                logging.error("You have asked to use the redis session storage backend, but "
                              "`redis` is not installed. Install it with `pip install redis`.")
                raise RuntimeError('redis package not installed')
            self._sessions = RedisSessionStorage()
        else:
            raise RuntimeError("Invalid session storage type")

        self._expiration = expiration

    def close(self):
        self._sessions.close()

    def set(self, session_id: str, data: SessionData):
        """
        Creates a new session or overwrite an existing session.

        Args:
            session_id (str):   The unique identifier for the session.
            data (SessionData): The initial data to store in the session.
        """
        session_data = SessionEntry(
            data=data if data is not None else {},
            expiration=time.time() + self._expiration,
        )
        self._sessions[session_id] = session_data

    def get_session(self, session_id: str) -> SessionData:
        """
        Retrieves a session by its ID.

        Args:
            session_id (str): The ID of the session to retrieve.

        Returns:
            SessionData: The session data if the session exists and is not expired, otherwise None.

        Raises:
            KeyError: If the session does not exist.
        """
        session: SessionEntry = self._sessions[session_id]
        current_time = time.time()
        if current_time < session.expiration:
            session = dc.replace(session, expiration=current_time + self._expiration)
            self._sessions[session_id] = session
            return session.data
        else:
            del self._sessions[session_id]
            raise KeyError('session does not exist')

    def delete_session(self, session_id: str) -> bool:
        """
        Deletes a session by its ID.

        Args:
            session_id (str): The ID of the session to delete.

        Returns:
            bool: True if the session was deleted,
                  False if the session did not exist.
        """
        if session_id in self._sessions:
            del self._sessions[session_id]
            return True
        return False

    def cleanup_expired_sessions(self) -> None:
        """
        Manually cleans up expired sessions from the store.
        """
        current_time: float = time.time()
        expired_session_ids: list[str] = [
            session_id
            for session_id, session in self._sessions.items()
            if current_time >= session.expiration
        ]
        for session_id in expired_session_ids:
            del self._sessions[session_id]

    def clear(self) -> None:
        """
        Clear all session data.
        """
        for session_id in self._sessions:
            del self._sessions[session_id]


class SessionWrapper(SessionData):
    """Wrap session with nice setter to auto-update session."""
    def __init__(self, key: str, data: SessionData, session_mgr: Session):
        self._key = key
        self._session_mgr = session_mgr
        self._data = data

    def __getitem__(self, name: str) -> SessionDataTypes:
        return self._data[name]

    def __setitem__(self, name: str, value: SessionDataTypes):
        self._data[name] = value
        self._session_mgr.set(self._key, self._data)

    def __delitem__(self, name: str):
        del self._data[name]
        self._session_mgr.set(self._key, self._data)

    def __iter__(self) -> Iterator:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)


class SessionMixin:
    """Mixin to access session data"""
    current_user: Union[str, None]

    def initialize(self, session: Session, **kwargs):
        self._session_mgr = session

    @property
    def session(self):
        if self.current_user:
            try:
                data = self._session_mgr.get_session(self.current_user)
            except KeyError:
                data = {}
                self._session_mgr.set(self.current_user, data)

            return SessionWrapper(
                self.current_user,
                data,
                self._session_mgr
            )
        else:
            return None



"""Sub-package __init__."""

from .arghandler import ArgumentHandler, ArgumentSource
from .decorators import (
    authenticated,
    catch_error,
    keycloak_role_auth,
    role_authorization,
    scope_role_auth,
    token_attribute_role_mapping_auth,
    TokenAttributeRoleMappingProtocol,
    validate_request,
)
from .handler import (
    KeycloakUsernameMixin,
    OpenIDCookieHandlerMixin,
    OpenIDLoginHandler,
    RestHandler,
    RestHandlerSetup,
)
from .server import RestServer

__all__ = [
    "RestServer",
    "RestHandlerSetup",
    "RestHandler",
    "KeycloakUsernameMixin",
    "OpenIDCookieHandlerMixin",
    "OpenIDLoginHandler",
    "authenticated",
    "catch_error",
    "role_authorization",
    "scope_role_auth",
    "keycloak_role_auth",
    "token_attribute_role_mapping_auth",
    "TokenAttributeRoleMappingProtocol",
    "ArgumentHandler",
    "ArgumentSource",
    "validate_request",
]



"""Utility functions for RestClient."""

import asyncio
import logging
from typing import Any, Optional

from .client import RestClient

LOGGER = logging.getLogger(__name__)


########################################################################################################################


try:
    import openapi_core
    from openapi_core.contrib import requests as openapi_core_requests
    from openapi_core.exceptions import OpenAPIError
except ImportError:
    pass  # if client code wants to use these features, then let the built-in errors raise


async def request_and_validate(
    rc: RestClient,
    openapi_spec: "openapi_core.OpenAPI",
    method: str,
    path: str,
    args: Optional[dict[str, Any]] = None,
) -> Any:
    """Make request and validate the response against a given OpenAPI spec.

    Useful for testing and debugging.

    NOTE: this essentially mimics RestClient.request() with added features.
    """
    url, kwargs = rc._prepare(method, path, args=args)

    # run request as async in case of other dependent, concurrent actions (ex: test suite runs server in same process)
    response = await asyncio.wrap_future(rc.session.request(method, url, **kwargs))  # type: ignore[var-annotated,arg-type]

    try:
        openapi_spec.validate_response(
            openapi_core_requests.RequestsOpenAPIRequest(response.request),
            openapi_core_requests.RequestsOpenAPIResponse(response),  # type: ignore[arg-type] # openapi uses protocols
        )
    except OpenAPIError as e:
        LOGGER.error(
            f"OpenAPI response validator encountered an error: '{e}'; more info below."
        )
        LOGGER.info(f"request: {vars(response.request)}")
        LOGGER.info(f"response: {vars(response)}")
        raise

    out = rc._decode(response.content)
    response.raise_for_status()
    return out


########################################################################################################################
